"""These SESSION commands are used to enter and exit the various
processors in the program.
"""


class ProcessorEntry:
    def aux2(self, **kwargs):
        """Enters the binary file dumping processor.

        APDL Command: /AUX2

        Notes
        -----
        Enters the binary file dumping processor (ANSYS auxiliary processor
        AUX2).  This processor is used to dump the contents of certain ANSYS
        binary files for visual examination.

        This command is valid only at the Begin Level.
        """
        command = "/AUX2,"
        return self.run(command, **kwargs)

    def aux3(self, **kwargs):
        """Enters the results file editing processor.

        APDL Command: /AUX3

        Notes
        -----
        Enters the results file editing processor (ANSYS auxiliary processor
        AUX3).  This processor is used to edit ANSYS results files.

        This command is valid only at the Begin Level.
        """
        command = "/AUX3,"
        return self.run(command, **kwargs)

    def aux12(self, **kwargs):
        """Enters the radiation processor.

        APDL Command: /AUX12

        Notes
        -----
        Enters the radiation processor (ANSYS auxiliary processor AUX12).  This
        processor supports the Radiation Matrix and the Radiosity Solver
        methods.

        This command is valid only at the Begin Level.
        """
        command = "/AUX12,"
        return self.run(command, **kwargs)

    def aux15(self, **kwargs):
        """Enters the IGES file transfer processor.

        APDL Command: /AUX15

        Notes
        -----
        Enters the IGES file transfer processor (ANSYS auxiliary processor
        AUX15), used to read an IGES data file into the ANSYS program.

        This command is valid only at the Begin Level.
        """
        command = "/AUX15,"
        return self.run(command, **kwargs)

    def finish(self, **kwargs):
        """Exits normally from a processor.

        APDL Command: FINISH

        Notes
        -----
        Exits any of the ANSYS processors or the DISPLAY program.  For the
        ANSYS processors, data will remain intact in the database but the
        database is not automatically written to a file (use the SAVE command
        to write the database to a file).  See also the /QUIT command for an
        alternate processor exit command.  If exiting POST1, POST26, or OPT,
        see additional notes below.

        POST1:  Data in the database will remain intact, including the POST1
        element table data, the path table data, the fatigue table data, and
        the load case pointers.

        POST26:  Data in the database will remain intact, except that POST26
        variables are erased and specification commands (such as FILE, PRTIME,
        NPRINT, etc.) are reset.  Use the /QUIT command to exit the processor
        and bypass these exceptions.

        This command is valid in any processor.  This command is not valid at
        the Begin level.
        """
        command = "FINISH,"
        return self.run(command, **kwargs)

    def map(self, kdim="", kout="", limit="", **kwargs):
        """Maps pressures from source points to target surface elements.

        APDL Command: MAP

        Parameters
        ----------
        kdim
            Interpolation key:

            0 or 2 - Interpolation is done on a surface (default).

            3 - Interpolation is done within a volume. This option is
                useful if the supplied source data is volumetric field
                data rather than surface data.

        kout
            Key to control how pressure is applied when a target node is
            outside of the source region:

            0 - Use the pressure(s) of the nearest source point for
                target nodes outside of the region (default).

            1 - Set pressures outside of the region to zero.

        limit
            Number of nearby points considered for interpolation. The
            minimum is 5; the default is 20. Lower values reduce
            processing time.  However, some distorted or irregular
            meshes will require a higher LIMIT value to find the
            points encompassing the target node in order to define the
            region for interpolation.

        Notes
        -----
        Maps pressures from source points to target surface elements.
        """
        return self.run(f"MAP,,{kdim},,{kout},{limit}", **kwargs)

    def post1(self, **kwargs):
        """Enters the database results postprocessor.

        APDL Command: /POST1

        Notes
        -----
        Enters the general database results postprocessor (POST1).  All load
        symbols (/PBC, /PSF, or /PBF) are automatically turned off with this
        command.

        This command is valid only at the Begin Level.
        """
        command = "/POST1,"
        return self.run(command, **kwargs)

    def post26(self, **kwargs):
        """Enters the time-history results postprocessor.

        APDL Command: /POST26

        Notes
        -----
        Enters the time-history results postprocessor (POST26).

        This command is valid only at the Begin Level.
        """
        command = "/POST26,"
        return self.run(command, **kwargs)

    def prep7(self, **kwargs):
        """Enters the model creation preprocessor.

        APDL Command: /PREP7

        Notes
        -----
        Enters the general input data preprocessor (PREP7).

        This command is valid only at the Begin Level.
        """
        command = "/PREP7,"
        return self.run(command, **kwargs)

    def quit(self, **kwargs):
        """Exits a processor.

        APDL Command: /QUIT

        Notes
        -----
        This is an alternative to the FINISH command.  If any cleanup or file
        writing is normally done by the FINISH command, it is bypassed if the
        /QUIT command is used instead.  A new processor may be entered after
        this command.  See the /EXIT command to terminate the run.

        This command is valid in any processor.  This command is not valid at
        the Begin level.
        """
        command = "/QUIT,"
        return self.run(command, **kwargs)

    def slashsolu(self, **kwargs):
        """Enters the solution processor.

        APDL Command: /SOLU

        Notes
        -----
        This command is valid only at the Begin Level.
        """
        command = "/SOLU,"
        return self.run(command, **kwargs)



"""loads a list of verification files
"""
import glob
import inspect
import os

module_path = os.path.dirname(inspect.getfile(inspect.currentframe()))


def load_vmfiles():
    """load vmfiles and store their filenames"""
    vmfiles = {}
    verif_path = os.path.join(module_path, "verif")
    for filename in glob.glob(os.path.join(verif_path, "*dat")):
        basename = os.path.basename(filename)
        vmname = os.path.splitext(basename)[0]
        vmfiles[vmname] = filename

    return vmfiles


# save the module from failing if the verification files are unavailable.
try:
    vmfiles = load_vmfiles()
except:
    vmfiles = []



from typing import Optional


class Status:
    def areas(self, **kwargs):
        """Specifies "Areas" as the subsequent status topic.

        APDL Command: AREAS

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"AREAS,"
        return self.run(command, **kwargs)

    def bool(self, **kwargs):
        """Specifies "Booleans" as the subsequent status topic.

        APDL Command: BOOL

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"BOOL,"
        return self.run(command, **kwargs)

    def ceqn(self, **kwargs):
        """Specifies "Constraint equations" as the subsequent status topic.

        APDL Command: CEQN

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"CEQN,"
        return self.run(command, **kwargs)

    def couple(self, **kwargs):
        """Specifies "Node coupling" as the subsequent status topic.

        APDL Command: COUPLE

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"COUPLE,"
        return self.run(command, **kwargs)

    def digit(self, **kwargs):
        """Specifies "Node digitizing" as the subsequent status topic.

        APDL Command: DIGIT

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"DIGIT,"
        return self.run(command, **kwargs)

    def elem(self, **kwargs) -> Optional[str]:
        """Specifies "Elements" as the subsequent status topic.

        APDL Command: ELEM

        Notes
        -----

        The STAT command should immediately follow this command,
        which should report the status for the specified topic.
        """
        command = f"ELEM,"
        return self.run(command, **kwargs)

    def etype(self, **kwargs) -> Optional[str]:
        """Specify "Element types" as the subsequent status topic.

        APDL Command: ETYPE

        This is a status [STAT] topic command.
        The STAT command should immediately follow this command,
        which should report the status for the specified topic.

        Examples
        --------
        >>> mapdl.et(1, 'SOLID186')
        >>> mapdl.etype()
        >>> mapdl.stat()
         ELEMENT TYPE        1 IS SOLID186     3-D 20-NODE STRUCTURAL SOLID
          KEYOPT( 1- 6)=        0      0      0        0      0      0
          KEYOPT( 7-12)=        0      0      0        0      0      0
          KEYOPT(13-18)=        0      0      0        0      0      0

        """
        return self.run("ETYPE", **kwargs)

    def febody(self, **kwargs):
        """Specifies "Body loads on elements" as the subsequent status topic.

        APDL Command: FEBODY

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"FEBODY,"
        return self.run(command, **kwargs)

    def fecons(self, **kwargs):
        """Specifies "Constraints on nodes" as the subsequent status topic.

        APDL Command: FECONS

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"FECONS,"
        return self.run(command, **kwargs)

    def fefor(self, **kwargs):
        """Specifies "Forces on nodes" as the subsequent status topic.

        APDL Command: FEFOR

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"FEFOR,"
        return self.run(command, **kwargs)

    def fesurf(self, **kwargs):
        """Specifies "Surface loads on elements" as the subsequent status topic.

        APDL Command: FESURF

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"FESURF,"
        return self.run(command, **kwargs)

    def geometry(self, **kwargs):
        """Specifies "Geometry" as the subsequent status topic.

        APDL Command: GEOMETRY

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"GEOMETRY,"
        return self.run(command, **kwargs)

    def keypts(self, **kwargs):
        """Specifies "Keypoints" as the subsequent status topic.

        APDL Command: KEYPTS

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"KEYPTS,"
        return self.run(command, **kwargs)

    def line(self, **kwargs):
        """Specifies "Lines" as the subsequent status topic.

        APDL Command: LINE

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"LINE,"
        return self.run(command, **kwargs)

    def mater(self, **kwargs):
        """Specifies "Material properties" as the subsequent status topic.

        APDL Command: MATER

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"MATER,"
        return self.run(command, **kwargs)

    def meshing(self, **kwargs):
        """Specifies "Meshing" as the subsequent status topic.

        APDL Command: MESHING

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"MESHING,"
        return self.run(command, **kwargs)

    def nodes(self, **kwargs):
        """Specifies "Nodes" as the subsequent status topic.

        APDL Command: NODES

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"NODES,"
        return self.run(command, **kwargs)

    def prim(self, **kwargs):
        """Specifies "Solid model primitives" as the subsequent status topic.

        APDL Command: PRIM

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"PRIM,"
        return self.run(command, **kwargs)

    def rcon(self, **kwargs):
        """Specifies "Real constants" as the subsequent status topic.

        APDL Command: RCON

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"RCON,"
        return self.run(command, **kwargs)

    def selm(self, **kwargs):
        """Specifies "Superelements" as the subsequent status topic.

        APDL Command: SELM

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands
        are generated by the GUI and will appear in the log file
        (Jobname.LOG) if status is requested for some items under
        Utility Menu> List> Status.  This command will be immediately
        followed by a STAT command, which will report the status for
        the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        return self.run("SELM,", **kwargs)

    def tble(self, **kwargs):
        """Specifies "Data table properties" as the subsequent status topic.

        APDL Command: TBLE

        Notes
        -----
        This is a status (STAT) topic command.  Status topic commands are
        generated by the GUI and will appear in the log file (Jobname.LOG) if
        status is requested for some items under Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"TBLE,"
        return self.run(command, **kwargs)

    def volumes(self, **kwargs):
        """Specifies "Volumes" as the subsequent status topic.

        APDL Command: VOLUMES

        Notes
        -----
        This is a status [STAT] topic command.  Status topic commands are
        generated by the GUI and appear in the log file (Jobname.LOG) if status
        is requested for some items by choosing Utility Menu> List> Status.
        This command will be immediately followed by a STAT command, which will
        report the status for the specified topic.

        If entered directly into the program, the STAT command should
        immediately follow this command.
        """
        command = f"VOLUMES,"
        return self.run(command, **kwargs)



"""Plotting helper for MAPDL using pyvista"""
import numpy as np

from ansys.mapdl.core import _HAS_PYVISTA
from ansys.mapdl.core.misc import get_bounding_box, unique_rows

from .theme import MapdlTheme

# Supported labels
BC_D = [
    "TEMP",
    "UX",
    "UY",
    "UZ",
    "VOLT",  # "MAG"
]
BC_F = [
    "HEAT",
    "FX",
    "FY",
    "FZ",
    "AMPS",
    "CHRGS",
    # "FLUX",
    "CSGZ",
]  # TODO: Add moments MX, MY, MZ
FIELDS = {
    "MECHANICAL": ["UX", "UY", "UZ", "FX", "FY", "FZ"],
    "THERMAL": ["TEMP", "HEAT"],
    "ELECTRICAL": ["VOLT", "CHRGS", "AMPS"],
}


# All boundary conditions:
BCS = BC_D.copy()
BCS.extend(BC_F)

# Allowed entities to plot their boundary conditions
ALLOWED_TARGETS = ["NODES"]


if _HAS_PYVISTA:

    import pyvista as pv

    # Symbols for constrains
    TEMP = pv.Sphere(center=(0, 0, 0), radius=0.5)

    UX = pv.Arrow(
        start=(-1, 0, 0), direction=(1, 0, 0), tip_length=1, tip_radius=0.5, scale=1.0
    )
    UY = pv.Arrow(
        start=(0, -1, 0), direction=(0, 1, 0), tip_length=1, tip_radius=0.5, scale=1.0
    )

    UZ = pv.Arrow(
        start=(0, 0, -1), direction=(0, 0, 1), tip_length=1, tip_radius=0.5, scale=1.0
    )

    FX = pv.Arrow(
        start=(-1, 0, 0),
        direction=(1, 0, 0),
        tip_length=0.5,
        tip_radius=0.25,
        scale=1.0,
    )
    FY = pv.Arrow(
        start=(0, -1, 0),
        direction=(0, 1, 0),
        tip_length=0.5,
        tip_radius=0.25,
        scale=1.0,
    )

    FZ = pv.Arrow(
        start=(0, 0, -1),
        direction=(0, 0, 1),
        tip_length=0.5,
        tip_radius=0.25,
        scale=1.0,
    )

    def get_VOLT():
        model_a = pv.Cylinder(
            center=(0, 0, 0), direction=(1, 0, 0), radius=0.2, height=2
        ).triangulate()

        model_b = pv.Cylinder(
            center=(0, 0, 0), direction=(0, 1, 0), radius=0.2, height=2
        ).triangulate()

        model_c = pv.Cylinder(
            center=(0, 0, 0), direction=(0, 0, 1), radius=0.2, height=2
        ).triangulate()

        result = model_a.merge(model_b).triangulate()
        result = result.merge(model_c)

        result.rotate_z(45.0, inplace=True)
        result.rotate_vector(
            vector=(1, -1, 0), angle=-45, point=(0, 0, 0), inplace=True
        )

        return result

    VOLT = get_VOLT()

    HEAT = pv.Cube(center=(0, 0, 0), x_length=1.0, y_length=1.0, z_length=1.0)

    BC_plot_settings = {
        "TEMP": {"color": "orange", "glyph": TEMP},
        "HEAT": {"color": "red", "glyph": HEAT},
        "UX": {"color": "red", "glyph": UX},
        "UY": {"color": "green", "glyph": UY},
        "UZ": {"color": "blue", "glyph": UZ},
        "VOLT": {"color": "yellow", "glyph": VOLT},
        "FX": {"color": "red", "glyph": FX},
        "FY": {"color": "green", "glyph": FY},
        "FZ": {"color": "blue", "glyph": FZ},
        "AMPS": {"color": "grey", "glyph": VOLT},
        "CHRGS": {"color": "grey", "glyph": VOLT},
    }

# Using * to force all the following arguments to be keyword only.
def _general_plotter(
    meshes,
    points,
    labels,
    *,
    cpos=None,
    show_bounds=False,
    show_axes=True,
    background=None,
    off_screen=None,
    notebook=None,
    # add_mesh kwargs:
    style=None,
    color="w",
    show_edges=None,
    edge_color=None,
    point_size=5.0,
    line_width=None,
    opacity=1.0,
    flip_scalars=False,
    lighting=None,
    n_colors=256,
    interpolate_before_map=True,
    cmap=None,
    render_points_as_spheres=False,
    render_lines_as_tubes=False,
    scalar_bar_args={},
    smooth_shading=None,
    feature_angle=30.0,
    show_scalar_bar=None,
    split_sharp_edges=None,
    # labels kwargs
    font_size=None,
    font_family=None,
    text_color=None,
    theme=None,
):
    """General pymapdl plotter for APDL geometry and meshes.

    Parameters
    ----------
    title : str, optional
        Add given title to plot.

    cpos : list(tuple(floats)), str
        The camera position to use.  You can either use a saved camera
        position or specify one of the following strings:

        - ``"xy"``
        - ``"xz"``
        - ``"yz"``
        - ``"yx"``
        - ``"zx"``
        - ``"zy"``
        - ``"iso"``

    off_screen : bool, optional
        Renders off screen when ``True``.  Useful for automated
        screenshots.

    window_size : list, optional
        Window size in pixels.  Defaults to ``[1024, 768]``

    notebook : bool, optional
        When True, the resulting plot is placed inline a jupyter
        notebook.  Assumes a jupyter console is active.  Automatically
        enables ``off_screen``.

    show_bounds : bool, optional
        Shows mesh bounds when ``True``.

    show_axes : bool, optional
        Shows a vtk axes widget.  Enabled by default.

    savefig : str, optional
        Saves screenshot to a file path. If used, ``notebook`` and
        ``off_screen`` are evaluated to ``False`` and ``True``
        respectively.

    style : string, optional
        Visualization style of the mesh.  One of the following:
        ``style='surface'``, ``style='wireframe'``,
        ``style='points'``.  Defaults to ``'surface'``. Note that
        ``'wireframe'`` only shows a wireframe of the outer geometry.

    color : string or 3 item list, optional
        Use to make the entire mesh have a single solid color.  Either
        a string, RGB list, or hex color string.  For example:
        ``color='white'``, ``color='w'``, ``color=[1, 1, 1]``, or
        ``color='#FFFFFF'``. Color will be overridden if scalars are
        specified.

    show_edges : bool, optional
        Shows the edges of a mesh.  Does not apply to a wireframe
        representation.

    edge_color : string or 3 item list, optional, defaults to black
        The solid color to give the edges when ``show_edges=True``.
        Either a string, RGB list, or hex color string.

    point_size : float, optional
        Point size of any nodes in the dataset plotted. Also applicable
        when ``style='points'``. Default ``5.0``

    line_width : float, optional
        Thickness of lines.  Only valid for wireframe and surface
        representations.  Default None.

    opacity : float, str, array-like
        Opacity of the mesh. If a single float value is given, it will be
        the global opacity of the mesh and uniformly applied everywhere
        should be between 0 and 1. A string can also be specified to map
        the scalars range to a predefined opacity transfer function
        (options include: ``'linear'``, ``'linear_r'``, ``'geom'``, or
        ``'geom_r'``).
        A string could also be used to map a scalars array from the mesh to
        the opacity (must have same number of elements as the
        ``scalars`` argument). Or you can pass a custom made transfer
        function that is an array either ``n_colors`` in length or shorter.

    n_colors : int, optional
        Number of colors to use when displaying scalars. Defaults to 256.
        The scalar bar will also have this many colors.

    cmap : str, list, optional
       Name of the Matplotlib colormap to us when mapping the
       ``scalars``.  See available Matplotlib colormaps.  Only
       applicable for when displaying ``scalars``. Requires Matplotlib
       to be installed.  ``colormap`` is also an accepted alias for
       this. If ``colorcet`` or ``cmocean`` are installed, their
       colormaps can be specified by name.

       You can also specify a list of colors to override an
       existing colormap with a custom one.  For example, to
       create a three color colormap you might specify
       ``['green', 'red', 'blue']``

    render_points_as_spheres : bool, optional
        Render points as spheres.

    render_lines_as_tubes : bool, optional
        Renders lines as tubes.

    smooth_shading : bool, optional
        Smoothly render curved surfaces when plotting.  Not helpful
        for all meshes.

    split_sharp_edges : bool, optional
        Split sharp edges exceeding 30 degrees when plotting with
        smooth shading.  Control the angle with the optional
        keyword argument ``feature_angle``.
        By default this is ``False``.

        .. note:: Note that enabling this will create a copy of
           the input mesh within the plotter.

    feature_angle : float, optional
        Angle to consider an edge a sharp edge. Default 30 degrees.

    theme : pyvista.DefaultTheme, optional
        PyVista theme. Defaults to `PyMAPDL theme <https://github
        .com/pyansys/pyansys-sphinx-theme>`_.

    Returns
    -------
    pyvista.Plotter
        Instance of ``pyvista.Plotter``.

    Examples
    --------
    Plot areas and modify the background color to ``'black'``

    >>> p = mapdl.aplot(background='black')

    Enable smooth_shading on an element plot.

    >>> p = mapdl.eplot(smooth_shading=True)

    Return the plotting instance, modify it, and display the plot.

    >>> pl = mapdl.aplot()
    >>> pl.show_bounds()
    >>> pl.set_background('black')
    >>> pl.add_text('my text')
    >>> pl.show()

    Save a screenshot to disk without showing the plot.

    >>> mapdl.eplot(background='w', show_edges=True, smooth_shading=True,
                    window_size=[1920, 1080], savefig='screenshot.png',
                    off_screen=True)

    """
    # Lazy import
    import pyvista as pv

    if theme is None:
        theme = MapdlTheme()

    pl = pv.Plotter(off_screen=off_screen, notebook=notebook, theme=theme)

    if background:
        pl.set_background(background)

    for point in points:
        pl.add_points(
            point["points"],
            scalars=point.get("scalars", None),
            color=color,
            show_edges=show_edges,
            edge_color=edge_color,
            point_size=point_size,
            line_width=line_width,
            opacity=opacity,
            flip_scalars=flip_scalars,
            lighting=lighting,
            n_colors=n_colors,
            interpolate_before_map=interpolate_before_map,
            cmap=cmap,
            render_points_as_spheres=render_points_as_spheres,
            render_lines_as_tubes=render_lines_as_tubes,
        )

    for mesh in meshes:
        pl.add_mesh(
            mesh["mesh"],
            scalars=mesh.get("scalars"),
            scalar_bar_args=scalar_bar_args,
            color=mesh.get("color", color),
            style=mesh.get("style", style),
            show_edges=show_edges,
            edge_color=edge_color,
            smooth_shading=smooth_shading,
            split_sharp_edges=split_sharp_edges,
            feature_angle=feature_angle,
            point_size=point_size,
            line_width=line_width,
            show_scalar_bar=show_scalar_bar,
            opacity=opacity,
            flip_scalars=flip_scalars,
            lighting=lighting,
            n_colors=n_colors,
            interpolate_before_map=interpolate_before_map,
            cmap=cmap,
            render_points_as_spheres=render_points_as_spheres,
            render_lines_as_tubes=render_lines_as_tubes,
        )

    for label in labels:
        # verify points are not duplicates
        points, idx, _ = unique_rows(np.array(label["points"]))
        labels = np.array(label["labels"])[idx].tolist()

        pl.add_point_labels(
            points,
            labels,
            show_points=False,
            shadow=False,
            font_size=font_size,
            font_family=font_family,
            text_color=text_color,
        )

    if cpos:
        pl.camera_position = cpos

    if show_bounds:
        pl.show_bounds()

    if show_axes:
        pl.show_axes()

    return pl


# Using * to force all the following arguments to be keyword only.
def general_plotter(
    meshes,
    points,
    labels,
    *,
    title="",
    cpos=None,
    show_bounds=False,
    show_axes=True,
    background=None,
    off_screen=None,
    savefig=None,
    window_size=None,
    notebook=None,
    # add_mesh kwargs:
    style=None,
    color="w",
    show_edges=None,
    edge_color=None,
    point_size=5.0,
    line_width=None,
    opacity=1.0,
    flip_scalars=False,
    lighting=None,
    n_colors=256,
    interpolate_before_map=True,
    cmap=None,
    render_points_as_spheres=False,
    render_lines_as_tubes=False,
    scalar_bar_args={},
    smooth_shading=None,
    show_scalar_bar=None,
    split_sharp_edges=None,
    # labels kwargs
    font_size=None,
    font_family=None,
    text_color=None,
    theme=None,
    return_plotter=False,
    return_cpos=False,
    mapdl=None,
    plot_bc=False,
    plot_bc_legend=None,
    plot_bc_labels=None,
    bc_labels=None,
    bc_target=None,
    bc_glyph_size=None,
    bc_labels_font_size=16,
):
    """General pymapdl plotter for APDL geometry and meshes.

    Parameters
    ----------
    title : str, optional
        Add given title to plot.

    cpos : list(tuple(floats)), str
        The camera position to use.  You can either use a saved camera
        position or specify one of the following strings:

        - ``"xy"``
        - ``"xz"``
        - ``"yz"``
        - ``"yx"``
        - ``"zx"``
        - ``"zy"``
        - ``"iso"``

    off_screen : bool, optional
        Renders off screen when ``True``.  Useful for automated
        screenshots.

    window_size : list, optional
        Window size in pixels.  Defaults to ``[1024, 768]``

    notebook : bool, optional
        When True, the resulting plot is placed inline a jupyter
        notebook.  Assumes a jupyter console is active.  Automatically
        enables off_screen.

    show_bounds : bool, optional
        Shows mesh bounds when ``True``.

    show_axes : bool, optional
        Shows a vtk axes widget.  Enabled by default.

    savefig : str, optional
        Saves screenshot to a file path.

    style : string, optional
        Visualization style of the mesh.  One of the following:
        ``style='surface'``, ``style='wireframe'``,
        ``style='points'``.  Defaults to ``'surface'``. Note that
        ``'wireframe'`` only shows a wireframe of the outer geometry.

    color : string or 3 item list, optional
        Use to make the entire mesh have a single solid color.  Either
        a string, RGB list, or hex color string.  For example:
        ``color='white'``, ``color='w'``, ``color=[1, 1, 1]``, or
        ``color='#FFFFFF'``. Color will be overridden if scalars are
        specified.

    show_edges : bool, optional
        Shows the edges of a mesh.  Does not apply to a wireframe
        representation.

    edge_color : string or 3 item list, optional,
        The solid color to give the edges when ``show_edges=True``.
        Either a string, RGB list, or hex color string.
        Defaults to black.

    point_size : float, optional
        Point size of any nodes in the dataset plotted. Also applicable
        when style='points'. Default ``5.0``

    line_width : float, optional
        Thickness of lines.  Only valid for wireframe and surface
        representations.  Default None.

    opacity : float, str, array-like
        Opacity of the mesh. If a single float value is given, it will be
        the global opacity of the mesh and uniformly applied everywhere -
        should be between 0 and 1. A string can also be specified to map
        the scalars range to a predefined opacity transfer function
        (options include: 'linear', 'linear_r', 'geom', 'geom_r').
        A string could also be used to map a scalars array from the mesh to
        the opacity (must have same number of elements as the
        ``scalars`` argument). Or you can pass a custom made transfer
        function that is an array either ``n_colors`` in length or shorter.

    n_colors : int, optional
        Number of colors to use when displaying scalars. Defaults to 256.
        The scalar bar will also have this many colors.

    cmap : str, list, optional
       Name of the Matplotlib colormap to us when mapping the
       ``scalars``.  See available Matplotlib colormaps.  Only
       applicable for when displaying ``scalars``. Requires Matplotlib
       to be installed.  ``colormap`` is also an accepted alias for
       this. If ``colorcet`` or ``cmocean`` are installed, their
       colormaps can be specified by name.

       You can also specify a list of colors to override an
       existing colormap with a custom one.  For example, to
       create a three color colormap you might specify
       ``['green', 'red', 'blue']``

    render_points_as_spheres : bool, optional
        Render points as spheres.

    render_lines_as_tubes : bool, optional
        Renders lines as tubes.

    smooth_shading : bool, optional
        Smoothly render curved surfaces when plotting.  Not helpful
        for all meshes.

    theme : pyvista.DefaultTheme, optional
        PyVista theme.  Defaults to PyMAPDL theme.

    return_plotter : bool, optional
        Return the plotting object rather than showing the plot and
        returning the camera position.  Default ``False``.
        This overrides the ``return_cpos`` value.

    return_cpos : bool, optional
        Returns the camera position as an array. Default ``False``.

    mapdl : Mapdl instance, optional
        If you want to use `plot_bc` keyword, the MAPDL instance
        needs to be passed as argument. Defaults to ``None``

    plot_bc : bool, optional
        Activate the plotting of the boundary conditions.
        Defaults to ``False``.

        .. warning:: This is in alpha state.

    plot_bc_legend : bool, optional
        Shows the boundary conditions legend.
        Defaults to ``False``

    plot_bc_labels : bool, optional
        Shows the boundary conditions label per node.
        Defaults to ``False``.

    bc_labels : List[str], Tuple(str), optional
        List or tuple of strings with the boundary conditions
        to plot, i.e. ["UX", "UZ"].
        You can obtain the allowed boundary conditions by
        evaluating ``ansys.mapdl.core.plotting.BCS``.
        You can use also the following shortcuts:

        * **'mechanical'**
          To plot the following mechanical boundary conditions:
          'UX', 'UY', 'UZ', 'FX', 'FY', and 'FZ'.
          Rotational or momentum boundary conditions are not
          allowed.

        * **'thermal'**
          To plot the following boundary conditions: 'TEMP' and
          'HEAT'.

        * **'electrical'**
          To plot the following electrical boundary conditions:
          'VOLT', 'CHRGS', and 'AMPS'.

        Defaults to all the allowed boundary conditions present
        in the responses of :func:`ansys.mapdl.core.Mapdl.dlist`
        and :func:`ansys.mapdl.core.Mapdl.flist()`.

    bc_target : List[str], Tuple(str), optional
        Specify the boundary conditions target
        to plot, i.e. "Nodes", "Elements".
        You can obtain the allowed boundary conditions target by
        evaluating ``ansys.mapdl.core.plotting.ALLOWED_TARGETS``.
        Defaults to only ``Nodes``.

    bc_glyph_size : float, optional
        Specify the size of the glyph used for the boundary
        conditions plotting.
        By default is ratio of the bounding box dimensions.

    bc_labels_font_size : float, optional
        Size of the text on the boundary conditions labels.
        By default it is 16.

    Returns
    -------
    cpos or pyvista.Plotter or None
        Camera position or instance of ``pyvista.Plotter`` or ``None`` depending
        on ``return_plotter`` and ``return_cpos``.

    Notes
    -----
    Plotting boundary conditions is still under-development, so feel free to share feedback
    or suggestion in `PyMAPDL <https://github.com/pyansys/pymapdl>`_.
    At the moment only nodal boundary conditions can be shown (``bc_target='Nodes'``), and only
    the following types of boundary conditions:

    +------------+--------------------------------------+
    | Field      | Boundary conditions                  |
    +============+======================================+
    | MECHANICAL | ["UX", "UY", "UZ", "FX", "FY", "FZ"] |
    +------------+--------------------------------------+
    | THERMAL    | ["TEMP", "HEAT"]                     |
    +------------+--------------------------------------+
    | ELECTRICAL | ["VOLT", "CHRGS", "AMPS"]            |
    +------------+--------------------------------------+

    Examples
    --------
    Plot areas and modify the background color to ``'black'``

    >>> cpos = mapdl.aplot(background='black')

    Enable smooth_shading on an element plot.

    >>> cpos = mapdl.eplot(smooth_shading=True)

    Plot boundary conditions "UX" and "UZ" on the nodes:

    >>> mapdl.nplot(plot_bc=True, bc_labels=["UX", "UZ"], plot_bc_labels=True)

    Return the plotting instance, modify it, and display the plot.

    >>> pl = mapdl.aplot(return_plotter=True)
    >>> pl.show_bounds()
    >>> pl.set_background('black')
    >>> pl.add_text('my text')
    >>> pl.show()

    Save a screenshot to disk without showing the plot.

    >>> mapdl.eplot(background='w', show_edges=True, smooth_shading=True,
                    window_size=[1920, 1080], savefig='screenshot.png',
                    off_screen=True)

    """
    if notebook:
        off_screen = True  # pragma: no cover

    if savefig:
        off_screen = True
        notebook = False

    # Getting the plotter
    pl = _general_plotter(
        meshes,
        points,
        labels,
        cpos=cpos,
        show_bounds=show_bounds,
        show_axes=show_axes,
        background=background,
        off_screen=off_screen,
        notebook=notebook,
        # add_mesh kwargs:
        style=style,
        color=color,
        show_edges=show_edges,
        edge_color=edge_color,
        point_size=point_size,
        line_width=line_width,
        opacity=opacity,
        flip_scalars=flip_scalars,
        lighting=lighting,
        n_colors=n_colors,
        interpolate_before_map=interpolate_before_map,
        cmap=cmap,
        render_points_as_spheres=render_points_as_spheres,
        render_lines_as_tubes=render_lines_as_tubes,
        scalar_bar_args=scalar_bar_args,
        smooth_shading=smooth_shading,
        show_scalar_bar=show_scalar_bar,
        split_sharp_edges=split_sharp_edges,
        # labels kwargs
        font_size=font_size,
        font_family=font_family,
        text_color=text_color,
        theme=theme,
    )

    if plot_bc:
        if not mapdl:
            raise ValueError(
                "An instance of `ansys.mapdl.core.mapdl._MapdlCore` "
                "should be passed using `mapdl` keyword if you are aiming "
                "to plot the boundary conditions (`plot_bc` is `True`)."
            )
        pl = bc_plotter(
            pl,
            mapdl=mapdl,
            plot_bc_legend=plot_bc_legend,
            plot_bc_labels=plot_bc_labels,
            bc_labels=bc_labels,
            bc_target=bc_target,
            bc_glyph_size=bc_glyph_size,
            bc_labels_font_size=bc_labels_font_size,
        )

    if title:  # Added here to avoid labels overlapping title
        pl.add_title(title, color=text_color)

    if return_cpos and return_plotter:
        raise ValueError(
            "'return_cpos' and 'return_plotter' cannot be both 'True' at the same time."
        )

    # permit user to save the figure as a screenshot
    if savefig:
        pl.show(title=title, auto_close=False, window_size=window_size, screenshot=True)
        pl.screenshot(savefig)

        # return unclosed plotter
        if return_plotter:
            return pl

        # if not returning plotter, close right away
        pl.close()

    else:
        if not return_plotter:
            pl.show()

    if return_plotter:
        return pl
    elif return_cpos:
        return pl.camera_position
    else:
        return None


def bc_plotter(
    pl,
    mapdl=None,
    bc_labels=None,
    bc_target=None,
    plot_bc_labels=False,
    plot_bc_legend=None,
    bc_glyph_size=None,
    bc_labels_font_size=16,
):

    if bc_labels:
        bc_labels = _bc_labels_checker(bc_labels)
    else:
        bc_labels = _bc_labels_default(mapdl)

    if bc_target:
        bc_target = _bc_target_checker(bc_target)
    else:
        bc_target = ["NODES"]  # bc_target_default()

    # We need to scale the glyphs, otherwise, they will be plot sized 1.
    # We are going to calculate the distance between the closest points,
    # so we will scaled to a percentage of this distance.
    # This might create very small points in cases there are a concentration of points.
    #
    # Later can find a way to plot them and keep their size constant independent of the zoom.

    if bc_glyph_size is None:
        bc_glyph_size = get_bounding_box(mapdl.mesh.nodes)
        bc_glyph_size = bc_glyph_size[bc_glyph_size != 0]

        if bc_glyph_size.size != 0:
            bc_glyph_size = bc_glyph_size.mean() * 0.75 / 10
        else:  # Case were there is only one node
            bc_glyph_size = 1

    if not isinstance(bc_glyph_size, (int, float)):
        raise ValueError("The 'bc_glyph_size' parameter can be only an int or float.")

    if "NODES" in bc_target:
        pl = bc_nodes_plotter(
            mapdl,
            pl,
            bc_labels,
            plot_bc_labels=plot_bc_labels,
            bc_glyph_size=bc_glyph_size,
            plot_bc_legend=plot_bc_legend,
            bc_labels_font_size=bc_labels_font_size,
        )

    # Add next things
    if "ELEM" in bc_target:
        pass

    return pl


def bc_nodes_plotter(
    mapdl,
    pl,
    bc_labels,
    plot_bc_labels=False,
    bc_glyph_size=1,
    plot_bc_legend=None,
    bc_labels_font_size=16,
):
    """Plot nodes BC given a list of labels."""
    nodes_xyz = mapdl.mesh.nodes
    nodes_num = mapdl.mesh.nnum

    bc_point_labels = None

    for each_label in bc_labels:
        if each_label in BC_D:
            bc = mapdl.get_nodal_constrains(each_label)

        elif each_label in BC_F:
            bc = mapdl.get_nodal_loads(each_label)

        else:
            raise Exception(f"The label '{each_label}' is not supported.")

        if bc.size == 0:  # There is no nodes with such label
            return pl

        bc_num = bc[:, 0].astype(int)
        bc_nodes = nodes_xyz[np.isin(nodes_num, bc_num), :]
        bc_values = bc[:, 1:].astype(float)
        bc_scale = abs(bc_values[:, 0])

        if bc_scale.max() != bc_scale.min():
            bc_scale = (bc_scale - bc_scale.min()) / (
                bc_scale.max() - bc_scale.min()
            ) + 0.5  # Normalization around 0.5
        else:
            bc_scale = (
                np.ones(bc_scale.shape) * 0.5
            )  # In case all the values are the same

        pcloud = pv.PolyData(bc_nodes)
        pcloud["scale"] = bc_scale * bc_glyph_size

        # Specify tolerance in terms of fraction of bounding box length.
        # Float value is between 0 and 1. Default is None.
        # If absolute is True then the tolerance can be an absolute distance.
        # If None, points merging as a preprocessing step is disabled.
        glyphs = pcloud.glyph(
            orient=False,
            scale="scale",
            # tolerance=0.05,
            geom=BC_plot_settings[each_label]["glyph"],
        )
        name_ = f"{each_label}"
        pl.add_mesh(
            glyphs,
            color=BC_plot_settings[each_label]["color"],
            style="surface",
            # style='wireframe',
            # edge_color=BC_plot_settings[each_label]['color'],
            # line_width=3,
            name=name_,
            label=name_,
        )

        if plot_bc_labels:
            if bc_point_labels is None:
                bc_point_labels = {each: "" for each in nodes_num}

            for id_, values in zip(bc_num, bc_values):
                if not bc_point_labels[id_]:
                    bc_point_labels[
                        id_
                    ] = f"Node: {id_}\n{each_label}: {values[0]:6.3f}, {values[1]:6.3f}"
                else:
                    bc_point_labels[
                        id_
                    ] = f"{bc_point_labels[id_]}\n{each_label}: {values[0]:6.3f}, {values[1]:6.3f}"

    if plot_bc_labels:
        pcloud = pv.PolyData(nodes_xyz)
        pcloud["labels"] = list(bc_point_labels.values())
        pl.add_point_labels(
            pcloud.points,
            pcloud["labels"],
            shape_opacity=0.25,
            font_size=bc_labels_font_size,
        )

    if plot_bc_legend:
        pl.add_legend(bcolor=None)

    return pl


def _bc_labels_checker(bc_labels):
    """Make sure we have allowed parameters and data types for ``bc_labels``"""
    if not isinstance(bc_labels, (str, list, tuple)):
        raise ValueError(
            "The parameter 'bc_labels' can be only a string, a list of strings or tuple of strings."
        )

    if isinstance(bc_labels, str):
        bc_labels = bc_labels.upper()
        if bc_labels not in BCS and bc_labels not in FIELDS:
            raise ValueError(
                f"The parameter '{bc_labels}' in 'bc_labels' is not supported.\n"
                "Please use any of the following:\n"
                f"{FIELDS}\nor any combination of:\n{BCS}"
            )
        bc_labels = [bc_labels]

    elif isinstance(bc_labels, (list, tuple)):
        if not all([isinstance(each, str) for each in bc_labels]):
            raise ValueError(
                "The parameter 'bc_labels' can be a list or tuple, but it should only contain strings inside them."
            )

        if not all([each.upper() in BCS for each in bc_labels]):
            raise ValueError(
                f"One or more parameters in 'bc_labels'({bc_labels}) are not supported.\n"
                f"Please use any combination of the following:\n{BCS}"
            )

        bc_labels = [each.upper() for each in set(bc_labels)]  # Removing duplicates too

    # Replacing field by equivalent fields.
    for each_field in FIELDS:
        if each_field in bc_labels:
            bc_labels.remove(each_field)
            bc_labels.extend(FIELDS[each_field])

    return bc_labels


def _bc_target_checker(bc_target):
    """Make sure we have allowed parameters and data types for ``bc_labels``"""
    if not isinstance(bc_target, (str, list, tuple)):
        raise ValueError(
            "The parameter 'bc_target' can be only a string, a list of strings or tuple of strings."
        )

    if isinstance(bc_target, str):
        if bc_target.upper() not in ALLOWED_TARGETS:
            raise ValueError(
                f"The parameter '{bc_target}' in 'bc_target' is not supported.\n"
                f"At the moments only the following are supported:\n{ALLOWED_TARGETS}"
            )
        bc_target = [bc_target.upper()]

    if isinstance(bc_target, (list, tuple)):
        if not all([each.upper() in ALLOWED_TARGETS for each in bc_target]):
            raise ValueError(
                "One or more parameters in 'bc_target' are not supported.\n"
                f"Please use any combination of the following:\n{ALLOWED_TARGETS}"
            )
        if not all([isinstance(each, str) for each in bc_target]):
            raise ValueError(
                "The parameter 'bc_target' can be a list or tuple, but it should only contain strings inside them."
            )

        bc_target = [each.upper() for each in bc_target]

    return bc_target


def _bc_labels_default(mapdl):
    """Get the labels from the MAPDL database."""
    flist = list(
        set(
            [
                each[1].upper()
                for each in mapdl.get_nodal_loads()
                if each[1].upper() in BCS
            ]
        )
    )
    dlist = list(
        set(
            [
                each[1].upper()
                for each in mapdl.get_nodal_constrains()
                if each[1].upper() in BCS
            ]
        )
    )
    return flist + dlist



"""
These PREP7 commands are used to read model data into the database,
list out the database, and control the numbering of entities in the
database.

"""


class Database:
    def aflist(self, **kwargs):
        """Lists the current data in the database.

        APDL Command: AFLIST

        Notes
        -----
        Lists the current data and specifications in the database.  If batch,
        lists all appropriate data.  If interactive, lists only summaries.
        """
        command = "AFLIST,"
        return self.run(command, **kwargs)

    def cdread(self, option="", fname="", ext="", fnamei="", exti="", **kwargs):
        """Reads a file of solid model and database information into the database.

        APDL Command: CDREAD

        Parameters
        ----------
        option
            Selects which data to read:

            ALL
                Read all geometry, material property, load, and
                component data (default).  Solid model geometry and loads
                will be read from the file ``Fnamei.Exti``.  All other data
                will be read from the file ``Fname.Ext``.

            DB
                Read all database information contained in file
                ``Fname.Ext``. This file should contain all information
                mentioned above except the solid model loads.
                If reading a ``.CDB`` file written with the ``GEOM`` option
                of the ``CDWRITE`` command, element types [``ET``] compatible with the
                connectivity of the elements on the file must be defined
                prior to reading.

            SOLID
                Read the solid model geometry and solid model
                loads from the file ``Fnamei.Exti``.
                This file could have been written by the ``CDWRITE`` or ``IGESOUT`` command.

            COMB
                Read the combined solid model and database
                information from the file ``Fname.Ext``.

        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).
            An unspecified directory path defaults to the working directory;
            in this case, you can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum). If there is an extension in
            ``fname``, this option is ignored.

        fnamei
            Name of the IGES file and its directory path (248 characters
            maximum, including directory). If you do not specify a directory
            path, it will default to your working directory and you can use all
            248 characters for the file name.

        exti
            Filename extension (eight-character maximum).

        Notes
        -----
        This command causes coded files of solid model (in IGES format) and
        database (in command format) information to be read.  These files are
        normally written by the ``CDWRITE`` or ``IGESOUT`` command.  Note that the
        active coordinate system in these files has been reset to Cartesian
        (``CSYS,0``).

        If a set of data exists prior to the ``CDREAD`` operation, that data set is
        offset upward to allow the new data to fit without overlap. The
        ``NOOFFSET`` command allows this offset to be ignored on a set-by-set
        basis, causing the existing data set to be overwritten with the new
        data set.

        When you write the geometry data using the ``CDWRITE,GEOM`` option, you use
        the ``CDREAD,DB`` option to read the geometry information.

        Using the ``CDREAD,COMB`` option will not write ``NUMOFF`` commands to offset
        entity ID numbers if there is no solid model in the database.

        Multiple CDB file imports cannot have elements with real constants in
        one file and section definitions in another. The section attributes
        will override the real constant attributes.  If you use ``CDREAD`` to
        import multiple CDB files, define all of the elements using only real
        constants, or using only section definitions.  Combining real constants
        and section definitions is not recommended.

        This command is valid in any processor.
        """
        return self.run(f"CDREAD,{option},{fname},{ext},,{fnamei},{exti}", **kwargs)

    def cdwrite(
        self, option="", fname="", ext="", fnamei="", exti="", fmat="", **kwargs
    ):
        """Writes geometry and load database items to a file.

        APDL Command: CDWRITE

        Parameters
        ----------
        option
            Selects which data to write:

            ALL - Write all appropriate geometry, material property,
                  load, and component data (default). Two files will
                  be produced. Fname.Ext will contain all data items
                  mentioned in "Notes", except the solid model
                  data. Fnamei.Exti will contain the solid model
                  geometry and solid model loads data in the form of
                  IGES commands. This option is not valid when
                  CDOPT,ANF is active.

            COMB - Write all data mentioned, but to a single file,
                   Fname.Ext. Solid model geometry data will be
                   written in either IGES or ANF format as specified
                   in the CDOPT command, followed by the remainder of
                   the data in the form of ANSYS commands. More
                   information on these (IGES/ANF) file formats is
                   provided in "Notes".

            DB - Write all database information except the solid model
                 and solid model loads to Fname.Ext in the form of
                 ANSYS commands. This option is not valid when
                 CDOPT,ANF is active.

            SOLID - Write only the solid model geometry and solid
                    model load data. This output will be in IGES or
                    ANF format, as specified in the CDOPT
                    command. More information on these (IGES/ANF) file
                    formats is provided in "Notes".

            GEOM - Write only element and nodal geometry data. Neither
                   solid model geometry nor element attribute data
                   will be written. One file, Fname.Ext, will be
                   produced. Use CDREAD,DB to read in a file written
                   with this option. Element types [ET] compatible
                   with the connectivity of the elements on the file
                   must first be defined before reading the file in
                   with CDREAD,DB.

            CM - Write only node and element component and geometry
            data to Fname.Ext.

            MAT - Write only material property data (both linear and
            nonlinear) to Fname.Ext.

            LOAD - Write only loads for current load step to
            Fname.Ext.

            SECT - Write only section data to Fname.Ext. Pretension
            sections are not included.

        fname
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).
            An unspecified directory path defaults to the working
            directory; in this case, you can use all 248 characters
            for the file name.

        ext
            Filename extension (eight-character maximum).

        fnamei
            Name of the IGES file and its directory path (248
            characters maximum, including directory). If you do not
            specify a directory path, it will default to your working
            directory and you can use all 248 characters for the file
            name.

        exti
            Filename extension (eight-character maximum).

        fmat
            Format of the output file (defaults to BLOCKED).

            BLOCKED - Blocked format. This format allows faster
                      reading of the output file. The time savings is
                      most significant when BLOCKED is used to read
                      .cdb files associated with very large models.

            UNBLOCKED - Unblocked format.

        Returns
        -------
        str
            Mapdl command output.

        Notes
        -----
        Load data includes the current load step only. Loads applied
        to the solid model (if any) are automatically transferred to
        the finite element model when this command is issued. CDWRITE
        writes out solid model loads for meshed models only. If the
        model is not meshed, the solid model loads cannot be
        saved. Component data include component definitions, but not
        assembly definitions. Appropriate NUMOFF commands are included
        at the beginning of the file; this is to avoid overlap of an
        existing database when the file is read in.

        Element order information (resulting from a WAVES command) is
        not written. The data in the database remain untouched.

        Solution control commands are typically not written to the
        file unless you specifically change a default solution
        setting.

        CDWRITE does not support the GSBDATA and GSGDATA commands, and
        these commands are not written to the file.

        The data may be reread (on a different machine, for example)
        with the CDREAD command. Caution: When the file is read in,
        the NUMOFF,MAT command may cause a mismatch between material
        definitions and material numbers referenced by certain loads
        and element real constants. See NUMOFF for details. Also, be
        aware that the files created by the CDWRITE command explicitly
        set the active coordinate system to Cartesian (CSYS,0).

        You should generally use the blocked format (Fmat = BLOCKED)
        when writing out model data with CDWRITE. This is a compressed
        data format that greatly reduces the time required to read
        large models through the CDREAD command. The blocked and
        unblocked formats are described in Chapter 3 of the Guide to
        Interfacing with ANSYS.

        If you use CDWRITE in any of the derived products (ANSYS Emag,
        ANSYS Professional), then before reading the file, you must
        edit the Jobname.cdb file to remove commands that are not
        available in the respective component product.

        The CDWRITE command writes PART information for any ANSYS
        LS-DYNA input file to the Jobname.cdb file via the EDPREAD
        command. (EDPREAD is not a documented command; it is written
        only when the CDWRITE command is issued.) The PART information
        can be automatically read in via the CDREAD command; however,
        if more than one Jobname.cdb file is read, the PART list from
        the last Jobname.cdb file overwrites the existing PART list of
        the total model. This behavior affects all PART-related
        commands contained in the Jobname.cdb file. You can join
        models, but not PART-related inputs, which you must modify
        using the newly-created PART numbers. In limited cases, an
        update of the PART list (EDWRITE,PUPDATE) is possible; doing
        so requires that no used combination of MAT/TYPE/REAL appears
        more than once in the list.

        The CDWRITE command does not support (for beam meshing) any
        line operation that relies on solid model associativity. For
        example, meshing the areas adjacent to the meshed line,
        plotting the line that contains the orientation nodes, or
        clearing the mesh from the line that contains orientation
        nodes may not work as expected. For more information about
        beam meshing, see Meshing Your Solid Model in the Modeling and
        Meshing Guide.

        IGES and ANF File Formats for Solid Model Geometry Information

        The format used for solid model geometry information is
        determined by the current CDOPT command setting. The default
        format is IGES.

        IGES option (default) to write solid model information (CDOPT,
        IGS):

        Before writing solid model entities, select all corresponding
        lower level entities (ALLSEL,BELOW,ALL).

        Examples
        --------
        Create a basic block and save it to disk.

        >>> mapdl.prep7()
        >>> mapdl.block(0, 1, 0, 1, 0, 1)
        >>> mapdl.et(1, 186)
        >>> mapdl.esize(0.25)
        >>> mapdl.vmesh('ALL')
        >>> mapdl.cdwrite('DB', '/tmp/mesh.cdb')
         TITLE =
         NUMBER OF ELEMENT TYPES =      1
                64 ELEMENTS CURRENTLY SELECTED.  MAX ELEMENT NUMBER =   64
               425 NODES CURRENTLY SELECTED.     MAX NODE NUMBER =     425
                 8 KEYPOINTS CURRENTLY SELECTED. MAX KEYPOINT NUMBER =   8
                12 LINES CURRENTLY SELECTED.     MAX LINE NUMBER =      12
                 6 AREAS CURRENTLY SELECTED.     MAX AREA NUMBER =       6
                 1 VOLUMES CURRENTLY SELECTED.   MAX VOL. NUMBER =       1
         WRITE ANSYS DATABASE AS AN ANSYS INPUT FILE: /tmp/mesh.cdb

        Optionally load the mesh into Python using the archive reader.

        >>> from ansys.mapdl import reader as pymapdl_reader
        >>> mesh = pymapdl_reader.Archive('/tmp/mesh.cdb')
        >>> mesh
        ANSYS Archive File mesh.cdb
         Number of Nodes:              425
         Number of Elements:           64
         Number of Element Types:      1
         Number of Node Components:    0
         Number of Element Components: 0

        """
        command = f"CDWRITE,{option},'{fname}',{ext},,{fnamei},{exti},{fmat}"
        return self.run(command, **kwargs)

    def cdopt(self, option="", **kwargs):
        """Specifies format to be used for archiving geometry.

        APDL Command: CDOPT

        Parameters
        ----------
        option
            IGES

            IGES - Write solid model geometry information using IGES format (default).

            ANF - Write solid model geometry information using ANSYS Neutral File format.

            STAT - Print out the current format setting.

        Notes
        -----
        This command controls your solid model geometry format for CDWRITE
        operations. The ANF option affects only the COMB and SOLID options of
        the CDWRITE command. All other options remain unaffected.

        This option setting is saved in the database.
        """
        command = "CDOPT,%s" % (str(option))
        return self.run(command, **kwargs)

    def cecheck(self, itemlab="", tolerance="", dof="", **kwargs):
        """Check constraint equations and couplings for rigid body motions.

        APDL Command: CECHECK

        Parameters
        ----------
        itemlab
            Item indicating what is to be checked:

            CE - Check constraint equations only

            CP - Check couplings only

            ALL - Check both CE and CP

        tolerance
            Allowed amount of out-of-balance for any constraint equation or
            coupled set. The default value of 1.0e-6 is usually good.

        dof
            Specifies which DOF is to be checked. Default is RIGID, the usual
            option. Other choices are individual DOF such as UX, ROTZ, etc. or
            THERM. The THERM option will check the constraint equations or
            coupled sets for free thermal expansions, whereas the individual
            DOFs check under rigid body motions. ALL is RIGID and THERM.

        Notes
        -----
        This command imposes a rigid body motion on the nodes attached to the
        constraint equation or coupled set and makes sure that no internal
        forces are generated for such rigid body motions. Generation of
        internal forces by rigid body motions usually indicates an error in the
        equation specification (possibly due to nodal coordinate rotations).
        The THERM option does a similar check to see that no internal forces
        are created by the equations if the body does a free thermal expansion
        (this check assumes a single isotropic coefficient of expansion).
        """
        command = "CECHECK,%s,%s,%s" % (str(itemlab), str(tolerance), str(dof))
        return self.run(command, **kwargs)

    def check(self, sele="", levl="", **kwargs):
        """Checks current database items for completeness.

        APDL Command: CHECK

        Parameters
        ----------
        sele
            Specifies which elements are to be checked:

            (blank) - Check all data.

            ESEL - Check only elements in the selected set and unselect any elements not producing
                   geometry check messages.  The remaining elements (those
                   producing check messages) can then be displayed and
                   corrected.  A null set results if no elements produce a
                   message.  Issue ESEL,ALL to select all elements before
                   proceeding.

        levl
            Used only with Sele = ESEL:

            WARN - Select elements producing warning and error messages.

            ERR - Select only elements producing error messages (default).

        Notes
        -----
        This command will not work if SHPP,OFF has been set. A similar,
        automatic check of all data is done before the solution begins.

        If the "Check Elements" option is invoked through the GUI (menu path
        Main Menu> Preprocessor> Meshing> Check Elems), the CHECK,ESEL logic is
        used to highlight elements in the following way:  good elements are
        blue, elements having warnings are yellow, and bad (error) elements are
        red.

        Note:: : The currently selected set of elements is not changed by this
        GUI function.

        This command is also valid in PREP7.
        """
        command = "CHECK,%s,%s" % (str(sele), str(levl))
        return self.run(command, **kwargs)

    def cncheck(
        self,
        option="",
        rid1="",
        rid2="",
        rinc="",
        intertype="",
        trlevel="",
        cgap="",
        cpen="",
        ioff="",
        **kwargs,
    ):
        """Provides and/or adjusts the initial status of contact pairs.

        APDL Command: CNCHECK

        Parameters
        ----------
        option
            Option to be performed:

            DETAIL - List all contact pair properties (default).

            SUMMARY - List only the open/closed status for each contact pair.

            POST - Execute a partial solution to write the initial
                   contact configuration to the Jobname.RCN file.

            ADJUST - Physically move contact nodes to the target in
                     order to close a gap or reduce penetration. The
                     initial adjustment is converted to structural
                     displacement values (UX, UY, UZ) and stored in
                     the Jobname.RCN file.

            RESET - Reset target element and contact element key
                    options and real constants to their default
                    values. This option is not valid for general
                    contact.

            AUTO - Automatically sets certain real constants and key
                   options to recommended values or settings in order
                   to achieve better convergence based on overall
                   contact pair behaviors. This option is not valid
                   for general contact.

            TRIM - Trim contact pair (remove certain contact and target elements).

            UNSE - Unselect certain contact and target elements.

        rid1, rid2, rinc
            For pair-based contact, the range of real constant pair
            ID's for which Option will be performed. If RID2 is not
            specified, it defaults to RID1. If no value is specified,
            all contact pairs in the selected set of elements are
            considered.

        intertype
            The type of contact interface (pair-based versus general
            contact) to be considered; or the type of contact pair to
            be trimmed/unselected/auto-set.

            (blank) - Include all contact definitions (pair-based and general contact).

            GCN - Include general contact definitions only (not valid
                  when Option = RESET or AUTO).

        trlevel
            Trimming level (used only when Option = TRIM or UNSE):

            (blank) - Normal trimming (default): remove/unselect
                      contact and target elements which are in
                      far-field.

            AGGRE - Aggressive trimming: remove/unselect contact and
                    target elements which are in far-field, and
                    certain elements in near-field.

        Notes
        -----
        The CNCHECK command provides information for
        surface-to-surface, node- to-surface, and line-to-line contact
        pairs (element types TARGE169, TARGE170, CONTA171, CONTA172,
        CONTA173, CONTA174, CONTA175, CONTA176, CONTA177). All contact
        and target elements of interest, along with the solid elements
        and nodes attached to them, must be selected for the command
        to function properly. For performance reasons, the program
        uses a subset of nodes and elements based on the specified
        contact regions (RID1, RID2, RINC) when executing the CNCHECK
        command.

        CNCHECK is available in both the PREP7 and SOLUTION
        processors, but only before the first solve operation (that
        is, only before the first load step or the first substep).

        If the contact and target elements were generated through mesh
        commands (AMESH, LMESH, etc.) instead of the ESURF command,
        you must issue MODMSH,DETACH before CNCHECK. Otherwise,
        CNCHECK will not work correctly.

        The following additional notes are available:

        The command CNCHECK,POST solves the initial contact configuration in
        one substep. After issuing this command, you can postprocess the
        contact result items as you would for any other converged load step;
        however, only the contact status, contact penetration or gap, and
        contact pressure will have meaningful values. Other contact quantities
        (friction stress, sliding distance, chattering) will be available but
        are not useful.

        Because Option = POST forces a solve operation, the PrepPost (PP)
        license does not work with CNCHECK,POST.
        """
        command = "CNCHECK,%s,%s,%s,%s,%s,%s,%s,%s,%s" % (
            str(option),
            str(rid1),
            str(rid2),
            str(rinc),
            str(intertype),
            str(trlevel),
            str(cgap),
            str(cpen),
            str(ioff),
        )
        return self.run(command, **kwargs)

    def igesout(self, fname="", ext="", att="", **kwargs):
        """Writes solid model data to a file in IGES Version 5.1 format.

        APDL Command: IGESOUT

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        att
            Attribute key:

            0 - Do not write assigned numbers and attributes of the
                solid model entities to the IGES file (default).

            1 - Write assigned numbers and attributes of solid model
                entities (keypoints, lines, areas, volumes) to the
                IGES file.  Attributes include MAT, TYPE, REAL, and
                ESYS specifications as well as associated solid model
                loads and meshing (keypoint element size, number of
                line divisions and spacing ratio) specifications.

        Notes
        -----
        Causes the selected solid model data to be written to a coded
        file in the IGES Version 5.1 format.  Previous data on this
        file, if any, are overwritten.  Keypoints that are not
        attached to any line are written to the output file as IGES
        entity 116 (Point).  Lines that are not attached to any area
        are written to the output file as either IGES Entity 100
        (Circular Arc), 110 (Line), or 126 (Rational B-Spline Curve)
        depending upon whether the ANSYS entity was defined as an arc,
        straight line, or spline.  Areas are written to the output
        file as IGES Entity 144 (Trimmed Parametric Surface).  Volumes
        are written to the output file as IGES entity 186 (Manifold
        Solid B-Rep Object).  Solid model entities to be written must
        have all corresponding lower level entities selected (use
        ALLSEL,BELOW,ALL) before issuing command.  Concatenated lines
        and areas are not written to the IGES file; however, the
        entities that make up these concatenated entities are written.

        Caution:: : Section properties assigned to areas, lines and
        other solid model entities will not be maintained when the
        model is exported using IGESOUT.

        If you issue the IGESOUT command after generating a beam mesh
        with orientation nodes, the orientation keypoints that were
        specified for the line (LATT) are no longer associated with
        the line and are not written out to the IGES file.  The line
        does not recognize that orientation keypoints were ever
        assigned to it, and the orientation keypoints do not "know"
        that they are orientation keypoints.  Thus the IGESOUT command
        does not support (for beam meshing) any line operation that
        relies on solid model associativity.  For example, meshing the
        areas adjacent to the meshed line, plotting the line that
        contains the orientation nodes, or clearing the mesh from the
        line that contains orientation nodes may not work as expected.
        See Meshing Your Solid Model in the Modeling and Meshing Guide
        for more information about beam meshing.
        """
        return self.run(f"IGESOUT,{fname},{ext},,{att}", **kwargs)

    def mfimport(self, fnumb="", option="", fname="", ext="", **kwargs):
        """Imports a new field into a current ANSYS Multi-field solver analysis.

        APDL Command: MFIMPORT

        Parameters
        ----------
        fnumb
            Field number specified by the MFELEM command.

        option
            Selects data to read.

            DB - Reads a CDB file. The CDB file name and extension are specified by Fname and
                 Ext.

        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        The field to be imported should be written to a CDB file (CDWRITE
        command). This file is read into the database, offsetting all existing
        element type numbers, node numbers, etc. in order to accommodate the
        imported field. (See the NUMOFF command for information on offset
        capabilities.) It then updates all of the previously issued MFxx
        commands to the new element type numbers. A new field is created using
        the specified field number, which must not currently exist. If there
        are no ANSYS Multi-field solver command files written for the existing
        fields in the database, one will be written for each field with the
        default name (see the MFCMMAND command). A MFCMMAND will be issued for
        the imported field as well.

        Repeat the MFIMPORT command to import additional fields.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = "MFIMPORT,%s,%s,%s,%s" % (
            str(fnumb),
            str(option),
            str(fname),
            str(ext),
        )
        return self.run(command, **kwargs)

    def nooffset(self, label="", **kwargs):
        """Prevents the CDREAD command from offsetting specified data items

        APDL Command: NOOFFSET

        Parameters
        ----------
        label
            Specifies items not to be offset.

            - ``"NODE"`` : Node numbers
            - ``"ELEM"`` : Element numbers
            - ``"KP"`` : Keypoint numbers
            - ``"LINE"`` : Line numbers
            - ``"AREA"`` : Area numbers
            - ``"VOLU"`` : Volume numbers
            - ``"MAT"`` : Material numbers
            - ``"TYPE"`` : Element type numbers
            - ``"REAL"`` : Real constant numbers
            - ``"CSYS"`` : Coordinate system numbers
            - ``"SECN"`` : Section numbers
            - ``"CP"`` : Coupled set numbers
            - ``"CE"`` : Constraint equation numbers
            - ``"CLEAR"`` : All items will be offset
            - ``"STATUS"`` : Shows which items are specified not to be offset.

        Notes
        -----
        The NOOFFSET command specifies data items not to be offset by a set of
        data read from a CDREAD command.
        """
        command = "NOOFFSET,%s" % (str(label))
        return self.run(command, **kwargs)

    def numcmp(self, label="", **kwargs):
        """Compresses the numbering of defined items.

        APDL Command: NUMCMP

        Parameters
        ----------
        label
            Items to be compressed:

            NODE - Node numbers

            ELEM - Element numbers

            KP - Keypoint numbers

            LINE - Line numbers

            AREA - Area numbers

            VOLU - Volume numbers

            MAT - Material numbers

            TYPE - Element type numbers

            REAL - Real constant numbers

            CP - Coupled set numbers

            SECN - Section numbers

            CE - Constraint equation numbers

            ALL - All item numbers

        Notes
        -----
        The NUMCMP command effectively compresses out unused item
        numbers by renumbering all the items, beginning with one and
        continuing throughout the model.  The renumbering order
        follows the initial item numbering order (that is, compression
        lowers the maximum number by "sliding" numbers down to take
        advantage of unused or skipped numbers).  All defined items
        are renumbered, regardless of whether or not they are actually
        used or selected.  Applicable related items are also checked
        for renumbering as described for the merge operation (NUMMRG).

        Compressing material numbers (NUMCMP,ALL or NUMCMP,MAT) does
        not update the material number referenced by either of the
        following:

        A temperature-dependent convection or surface-to-surface
        radiation load (SF, SFE, SFL, SFA)

        Real constants for multi-material elements (such as SOLID65)

        Compression is usually not required unless memory space is limited and
        there are large gaps in the numbering sequence.
        """
        command = "NUMCMP,%s" % (str(label))
        return self.run(command, **kwargs)

    def nummrg(self, label="", toler="", gtoler="", action="", switch="", **kwargs):
        """Merges coincident or equivalently defined items.

        APDL Command: NUMMRG

        Parameters
        ----------
        label
            Items to be merged:

            NODE - Nodes

            ELEM - Elements

            KP - Keypoints (will also merge lines, areas, and volumes)

            MAT - Materials

            TYPE - Element types

            REAL - Real constants

            CP - Coupled sets

            CE - Constraint equations

            ALL - All items

        toler
            Range of coincidence.  For Label = NODE and KP, defaults to 1.0E-4
            (based on maximum Cartesian coordinate difference between nodes or
            keypoints).  For Label = MAT, REAL, and CE, defaults to 1.0E-7
            (based on difference of the values normalized by the values).  Only
            items within range are merged.  (For keypoints attached to lines,
            further restrictions apply.  See the GTOLER field and Merging Solid
            Model Entities below.)

        gtoler
            Global solid model tolerance -- used only when merging keypoints
            attached to lines.  If specified, GTOLER will override the internal
            relative solid model tolerance.  See Merging Solid Model Entities
            below.

        action
            Specifies whether to merge or select coincident items.

            SELE - Select coincident items but do not merge. Action = SELE is only valid for Label
                   = NODE.

            (Blank) - Merge the coincident items (default).

        switch
            Specifies whether the lowest or highest numbered coincident item is
            retained after the merging operation.  This option does not apply
            to keypoints; i.e., for Label = KP, the lowest numbered keypoint is
            retained regardless of the Switch setting.

            LOW - Retain the lowest numbered coincident item after the merging operation
                  (default).

            HIGH - Retain the highest numbered coincident item after the merging operation.

        Notes
        -----
        After issuing the command, the area and volume sizes (ASUM and VSUM)
        may give slightly different results. In order to obtain the same
        results as before, use /FACET, /NORMAL, and ASUM / VSUM.

        The merge operation is useful for tying separate, but coincident, parts
        of a model together. If not all items are to be checked for merging,
        use the select commands (NSEL, ESEL, etc.) to select items.  Only
        selected items are included in the merge operation for nodes,
        keypoints, and elements.

        By default, the merge operation retains the lowest numbered coincident
        item.  Higher numbered coincident items are deleted.  Set Switch to
        HIGH to retain the highest numbered coincident item after the merging
        operation.  Applicable related items are also checked for deleted item
        numbers and if found, are replaced with the retained item number.  For
        example, if nodes are merged, element connectivities (except
        superelements), mesh item range associativity, coupled degrees of
        freedom, constraint equations, master degrees of freedom, gap
        conditions, degree of freedom constraints, nodal force loads, nodal
        surface loads, and nodal body force loads are checked.  Merging
        material numbers [NUMMRG,ALL or NUMMRG,MAT] does not update the
        material number referenced:

        By temperature-dependent film coefficients as part of convection load
        or a temperature-dependent emissivity as part of a surface-to-surface
        radiation load [SF, SFE, SFL, SFA]

        By real constants for multi-material elements (such as SOLID65)

        If a unique load is defined among merged nodes, the value is kept and
        applied to the retained node.  If loads are not unique (not
        recommended), only the value on the lowest node (or highest if Switch =
        HIGH) will be kept, except for "force" loads for which the values will
        be summed if they are not defined using tabular boundary conditions.

        Note:: The unused nodes (not recommended) in elements, couplings,
        constraint equations, etc. may become active after the merge operation.

        The Action field provides the option of visualizing the coincident
        items before the merging operation.

        Caution:: When merging entities in a model that has already been
        meshed, the order in which you issue multiple NUMMRG commands is
        significant.  If you want to merge two adjacent meshed regions that
        have coincident nodes and keypoints, always merge nodes [NUMMRG,NODE]
        before merging keypoints [NUMMRG,KP].  Merging keypoints before nodes
        can result in some of the nodes becoming "orphaned"; that is, the nodes
        lose their association with the solid model.  Orphaned nodes can cause
        certain operations (such as boundary condition transfers, surface load
        transfers, and so on) to fail. However, using NUMMRG should be avoided
        if at all possible, as the procedure outlined above may even cause
        meshing failure, especially after multiple merging and meshing
        operations.

        After a NUMMRG,NODE, is issued, some nodes may be attached to more than
        one solid entity. As a result, subsequent attempts to transfer solid
        model loads to the elements may not be successful. Issue NUMMRG,KP to
        correct this problem. Do NOT issue VCLEAR before issuing NUMMRG,KP.

        For NUMMRG,ELEM, elements must be identical in all aspects, including
        the direction of the element coordinate system.

        For certain solid and shell elements (181, 185, 190, etc) ANSYS will
        interpret coincident faces as internal and eliminate them. To prevent
        this from occurring, shrink the entities by a very small factor to
        delineate coincident items (/SHRINK, 0.0001) and no internal nodes,
        lines, areas or elements will be eliminated.

        When working with solid models, you may have better success with the
        gluing operations (AGLUE, LGLUE, VGLUE). Please read the following
        information when attempting to merge solid model entities.

        Gluing Operations vs. Merging Operations

        Adjacent, touching regions can be joined by gluing them (AGLUE, LGLUE,
        VGLUE) or by merging coincident keypoints (NUMMRG,KP, which also causes
        merging of identical lines, areas, and volumes).  In many situations,
        either approach will work just fine. Some factors, however, may lead to
        a preference for one method over the other.

        Geometric Configuration

        Gluing is possible regardless of the initial alignment or offset of the
        input entities. Keypoint merging is  possible only if each keypoint on
        one side of the face to be joined is matched by a coincident keypoint
        on the other side. This is commonly the case after a symmetry
        reflection (ARSYM or VSYMM) or a copy (AGEN or VGEN),  especially for a
        model built entirely in ANSYS rather than imported from a CAD system.
        When the geometry is  extremely precise, and the configuration is
        correct for  keypoint merging, NUMMRG is more efficient and robust than
        AGLUE or VGLUE.
        """
        command = "NUMMRG,%s,%s,%s,%s,%s" % (
            str(label),
            str(toler),
            str(gtoler),
            str(action),
            str(switch),
        )
        return self.run(command, **kwargs)

    def numoff(self, label="", value="", **kwargs):
        """Adds a number offset to defined items.

        APDL Command: NUMOFF

        Parameters
        ----------
        label
            Apply offset number to one of the following sets of items:

            NODE - Nodes

            ELEM - Elements

            KP - Keypoints

            LINE - Lines

            AREA - Areas

            VOLU - Volumes

            MAT - Materials

            TYPE - Element types

            REAL - Real constants

            CP - Coupled sets

            SECN - Section numbers

            CE - Constraint equations

            CSYS - Coordinate systems

        value
            Offset number value (cannot be negative).

        Notes
        -----
        Useful for offsetting current model data to prevent overlap if another
        model is read in. CDWRITE automatically writes the appropriate NUMOFF
        commands followed by the model data to File.CDB.  Therefore, when the
        file is read, any model already existing in the database is offset
        before the model data on the file is read.

        Offsetting material numbers with this command  [NUMOFF,MAT] does not
        update the material number referenced by either of the following:

        A temperature-dependent convection or surface-to-surface radiation load
        [SF, SFE, SFL, SFA]

        Real constants for multi-material elements (such as SOLID65).

        Therefore, a mismatch may exist between the material definitions and
        the material numbers referenced.
        """
        command = "NUMOFF,%s,%s" % (str(label), str(value))
        return self.run(command, **kwargs)

    def numstr(self, label="", value="", **kwargs):
        """Establishes starting numbers for automatically numbered items.

        APDL Command: NUMSTR

        Parameters
        ----------
        label
            Apply starting number to one of the following sets of items:

            NODE - Node numbers.  Value defaults (and is continually reset) to 1 + maximum node
                   number in model.  Cannot be reset lower.

            ELEM - Element numbers.  Value defaults (and is continually reset) to 1 + maximum
                   element number in model.  Cannot be reset lower.

            KP - Keypoint numbers.  Value defaults to 1.  Only undefined numbers are used.
                 Existing keypoints are not overwritten.

            LINE - Line numbers.  Value defaults to 1.  Only undefined numbers are used.  Existing
                   lines are not overwritten.

            AREA - Area numbers.  Value defaults to 1.  Only undefined numbers are used.  Existing
                   areas are not overwritten.

            VOLU - Volume numbers.  Value defaults to 1.  Only undefined numbers are used.
                   Existing volumes are not overwritten.

            DEFA - Default.  Returns all starting numbers to their default values.

        value
            Starting number value.

        Notes
        -----
        Establishes starting numbers for various items that may have numbers
        automatically assigned (such as element numbers with the EGEN command,
        and node and solid model entity numbers with the mesh [AMESH, VMESH,
        etc.] commands).  Use NUMSTR,STAT to display settings.  Use NUMSTR,DEFA
        to reset all specifications back to defaults.  Defaults may be lowered
        by deleting and compressing items (i.e., NDELE and NUMCMP,NODE for
        nodes, etc.).

        Note:: : A mesh clear operation (VCLEAR, ACLEAR, LCLEAR, and KCLEAR)
        automatically sets starting node and element numbers to the highest
        unused numbers.  If a specific starting node or element number is
        desired, issue NUMSTR after the clear operation.
        """
        command = "NUMSTR,%s,%s" % (str(label), str(value))
        return self.run(command, **kwargs)



class Aux15:
    def igesin(self, fname="", ext="", **kwargs):
        """Transfers IGES data from a file into MAPDL.

        APDL Command: IGESIN

        Parameters
        ----------
        fname : str
            File name and directory path (248 characters maximum,
            including the characters needed for the directory path).  An
            unspecified directory path defaults to the working directory;
            in this case, you can use all 248 characters for the file
            name.  The file name defaults to JOBNAME

        ext : str
            Filename extension (eight-character maximum).  This is
            optional and the extension can be included in the ``fname``
            parameter.  The extension defaults to CAD if ``fname`` is
            blank.

        Notes
        -----
        Reads a file containing IGES data and transfers it into the ANSYS
        database.  The file transferred is the IGES Version 5.1, ASCII
        format file.  IGES (Initial Graphics Exchange Specification) is a
        neutral format developed by the U.S. Dept. of Commerce, National
        Institute of Standards and Technology.  There is no output
        transfer file written since the transferred data is read directly
        into the MAPDL database.

        You can import multiple files into a single database, but you must
        use the same import option (set with the IOPTN command) for each
        file.

        The IOPTN command sets the parameters for reading the file.  Files
        read via the SMOOTH method (the only available method) use the
        standard database.

        Examples
        --------
        Read in an IGES file in the current MAPDL directory.

        >>> mapdl.igesin('bracket.iges')

        Read in an example IGES file

        >>> from ansys.mapdl.core import examples
        >>> bracket_file = examples.download_bracket()
        >>> mapdl.igesin(bracket_file)

        Read in a igesfile from a full path

        >>> mapdl.igesin('C:\\Users\\user\\bracket.iges)

        """
        return self.run(f"IGESIN,{fname},{ext}", **kwargs)

    def ioptn(self, lab="", val1="", **kwargs):
        """Controls options relating to importing a model.

        APDL Command: IOPTN

        Parameters
        ----------
        lab
            Label identifying the import option.  The meaning of VAL1 varies
            depending on Lab.

            STAT - List overall status of import facilities, including current option values.
                   VAL1 is ignored.

            DEFA - Set default values for all import options.  VAL1is ignored.

            MERG - Entity merge option.  VAL1 can be:

            YES - Automatic merging is performed (default).

            NO - No merging of entities.

            SOLID - Solid option.  VAL1 can be:

            YES - Solid is created automatically (default).

            NO - No solid created.

            GTOLER - Entity merging tolerance. If IGES = SMOOTH, the GTOLER,VAL1 can be:

            DEFA - Use system defaults (default).

            FILE - Use tolerance from the imported file.

            n - A user-specified tolerance value.

            IGES - IGES import option.  VAL1 can be:

            STAT - List status of IGES related options in the output window.

            SMOOTH (or RV52) - Use more robust IGES revision 5.2 import function (default).

            SMALL - Small areas option.   VAL1 can be:

            YES - Small areas are deleted (default).

            NO - Small areas are retained.

        val1
            Additional input value as described under each Lab option.

        Notes
        -----
        Controls various options during a model file transfer.  A global solid
        model tolerance (GTOLER) can be specified.

        The SMALL,YES option (default) delete small areas  and can cause
        geometrical inconsistencies that could cause the import process to
        abort.  Retaining the small areas increases processor time and memory
        usage.

        The data is stored in the standard ANSYS graphics database.

        The IGES,SMOOTH (default) option is capable of reading in any rational
        B-spline curve entity (type 126), or rational B-spline surface entity
        (type 128) with a degree less than or equal to 20.  Attempts to read in
        B-spline curve or surface entities of degree higher than 20 may result
        in error messages.

        If you issue the /CLEAR command, the IOPTN settings return to their
        defaults.

        For MERG,YES, merging of coincident geometry items is performed
        automatically when the IGESIN command is issued (that is, an internal
        NUMMRG,KP command is issued).  The model is merged with the
        consideration tolerance (TOLER on NUMMRG) set equal to 0.75 * the
        shortest distance between the endpoints of any active line. See the
        NUMMRG command for more information about the tolerances. In most
        cases, the default merging is appropriate.  Use the IOPTN command when
        you want to:

        Disable merging operations.

        Override the default merging and specify a global solid model tolerance
        value (GTOLER).

        Disable the automatic creation of solids (SOLID).

        The IOPTN command should be issued before the IGESIN command. You
        cannot change these options after your model has been imported or
        created. If you must change the options:

        Clear the database (/CLEAR) or exit and restart the program.

        Set the correct options.

        Reimport or recreate the model.

        This command is valid in any processor.
        """
        command = f"IOPTN,{lab},{val1}"
        return self.run(command, **kwargs)



class Setup:
    def rmresume(self, fname="", ext="", **kwargs):
        """Resumes ROM data from a file.

        APDL Command: RMRESUME

        Parameters
        ----------
        fname
            Name and directory path of the ROM database file (248 character
            maximum). Default to Jobname.

        ext
            Extension of the ROM database file. Default to .rom.

        Notes
        -----
        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"RMRESUME,{fname},{ext}"
        return self.run(command, **kwargs)

    def rmsave(self, fname="", ext="", **kwargs):
        """Saves ROM data to file.

        APDL Command: RMSAVE

        Parameters
        ----------
        fname
            Name and directory path of the ROM database file. Default to
            Jobname.

        ext
            Extension of the ROM database file. Default to .rom.

        Notes
        -----
        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        return self.run(f"RMSAVE,{fname},{ext}", **kwargs)



"""Contains the MapdlDb classes, allowing the access to MAPDL DB from Python."""
from enum import Enum
from functools import wraps
import os
import time
from warnings import warn
import weakref

try:
    from ansys.api.mapdl.v0 import mapdl_db_pb2_grpc
except ImportError:  # pragma: no cover
    raise ImportError(
        "Please upgrade the 'ansys.api.mapdl' package to at least v0.5.1."
        "You can use 'pip install ansys-api-mapdl --upgrade"
    )

import grpc

from ..mapdl_grpc import MapdlGrpc


class WithinBeginLevel:

    """Context manager to run MAPDL within the being level."""

    def __init__(self, mapdl):
        """Initialize this context manager."""
        self._mapdl = mapdl

    def __enter__(self):
        """Enter the begin level and cache the current routine."""
        self._mapdl._cache_routine()
        if "BEGIN" not in self._mapdl._cached_routine.upper():
            self._mapdl.finish()

    def __exit__(self, *args, **kwargs):
        """Exit the begin level and reload the previous routine."""
        if "BEGIN" not in self._mapdl._cached_routine.upper():
            self._mapdl._resume_routine()


def check_mapdl_db_is_alive(function):
    """
    Decorator to check that the MAPDL.DB has started.

    It works for the DB object (DBDef) and for the derived object which has "_db" attribute.
    """

    @wraps(function)
    def wrapper(self, *args, **kwargs):
        if hasattr(self, "active"):
            active = self.active
        elif hasattr(self, "_db"):
            active = self._db.active
        else:  # pragma: no cover
            raise Exception("The DB object could not be found.")

        if not active:
            self._mapdl._log.error(
                f"Please start the MAPDL DB Server to access '{function.__name__}'."
            )
            return None
        return function(self, *args, **kwargs)

    return wrapper


class DBDef(Enum):  # From MAPDL ansysdef.inc include file

    """Database type definitions."""

    DB_SELECTED = 1
    DB_NUMDEFINED = 12
    DB_NUMSELECTED = 13
    DB_MAXDEFINED = 14
    DB_MAXRECLENG = 15
    DB_GETNEXTRECD = 16
    DB_MAXALLOC = 17
    DB_OBJFLAG = -777
    DB_NEXT = -888
    DB_NEXT_ALLOC = -889
    DB_MODIFIED = -999
    DB_INTERNAL = -9999


class MapdlDb:

    """
    Abstract mapdl database class.  Created from a ``Mapdl`` instance.

    Examples
    --------
    Create a nodes instance.

    >>> from ansys.mapdl.core import launch_mapdl
    >>> mapdl = launch_mapdl()
    >>> # create nodes...
    >>> nodes = mapdl.db.nodes
    >>> print(nodes)
    MAPDL Database Nodes
        Number of nodes:          270641
        Number of selected nodes: 270641
        Maximum node number:      270641

    >>> mapdl.nsel("NONE")
    >>> print(nodes)
    MAPDL Database Nodes
        Number of nodes:          270641
        Number of selected nodes: 0
        Maximum node number:      270641

    Return the selection status and the coordinates of node 22.

    >>> nodes = mapdl.db.nodes
    >>> sel, coord = nodes.coord(22)
    >>> coord
    (1.0, 0.5, 0.0, 0.0, 0.0, 0.0)

    """

    def __init__(self, mapdl):
        """Initialize this class."""
        if not isinstance(mapdl, MapdlGrpc):  # pragma: no cover
            raise TypeError("``mapdl`` must be a MapdlGrpc instance")
        self._mapdl_weakref = weakref.ref(mapdl)
        self._stub = None
        self._channel = None
        self._ip = None
        self._server = {}
        self._channel_str = None
        self._state = None
        self._nodes = None
        self._elems = None

    def __str__(self):
        """Return the string representation of this class."""
        if self.active:
            return f"MAPDL database server active at {self._channel_str}"
        return f"MAPDL database server not active"

    @property
    def _mapdl(self):
        """Return the weakly referenced instance of mapdl."""
        return self._mapdl_weakref()

    def _start(self) -> int:
        """
        Lower level start of the database server.

        Returns
        -------
        int
            Port of the database server.

        """
        self._mapdl._log.debug("Starting MAPDL server")

        # database server must be run from the "BEGIN" level
        self._mapdl._cache_routine()
        with WithinBeginLevel(self._mapdl):
            self._mapdl.run("/DBS,SERVER,START")

        # Scan the DBServer.info file to get the Port Number

        # Default is 50055
        # wait for start
        tstart = time.time()
        timeout = 1
        status = self._mapdl._download_as_raw("DBServer.info").decode()
        while status == "":  # pragma: no cover
            status = self._mapdl._download_as_raw("DBServer.info").decode()
            time.sleep(0.05)
            if time.time() - tstart > timeout:
                raise TimeoutError(
                    f"Unable to start database server in {timeout} second(s)"
                )

        try:
            # expected of the form 'Port : 50055'
            port = int(status.split(":")[1])
        except Exception as e:  # pragma: no cover
            self._mapdl._log.error(
                "Unable to read port number from '%s' due to\n%s", status, str(e)
            )
            port = 50055

        self._mapdl._log.debug("MAPDL database server started on port %d", port)
        return port

    @property
    def active(self) -> bool:
        """Return if the database server is active."""
        return "NOT" not in self._status()

    def start(self, timeout=10):
        """
        Start the gRPC MAPDL database server.

        Parameters
        ----------
        timeout : float, optional
            Timeout to start the service.

        Examples
        --------
        >>> mapdl.db.start()
        """

        # only start if not already running
        is_running = self.active
        self._mapdl._log.debug("MAPDL DB server running: %s", str(is_running))
        if is_running:
            return
        db_port = self._start()

        self._ip = self._mapdl._ip

        # permit overriding db_port via env var for CI
        if "PYMAPDL_DB_PORT" in os.environ:
            db_port_str = os.environ.get("PYMAPDL_DB_PORT")
            try:
                db_port = int(db_port_str)
            except ValueError:  # pragma: no cover
                raise ValueError(
                    f"Invalid port '{db_port_str}' specified in the env var PYMAPDL_DB_PORT"
                )

        self._server = {"ip": self._ip, "port": db_port}
        self._channel_str = f"{self._ip}:{db_port}"

        self._channel = grpc.insecure_channel(self._channel_str)
        self._state = grpc.channel_ready_future(self._channel)
        self._stub = mapdl_db_pb2_grpc.MapdlDbServiceStub(self._channel)

        # wait until the channel matures
        tstart = time.time()
        while ((time.time() - tstart) < timeout) and not self._state._matured:
            time.sleep(0.01)

        if not self._state._matured:  # pragma: no cover
            raise RuntimeError(
                "Unable to establish connection to MAPDL database server"
            )
        self._mapdl._log.debug("Established connection to MAPDL database server")

    def _stop(self):
        """Stop the MAPDL database service."""
        with WithinBeginLevel(self._mapdl):
            return self._mapdl.run("/DBS,SERVER,STOP")

    def stop(self):
        """
        Shutdown the MAPDL database service and close the connection.

        Examples
        --------
        Stop the database service.

        >>> mapdl.db.stop()
        """
        if not self.active:
            warn("Server is already shutdown")
            return

        self._mapdl._log.debug("Closing the connection with the MAPDL DB Server")
        self._stop()
        self._channel.close()
        self._channel = None
        self._stub = None
        self._state = None

    def _status(self):
        """
        Return the status of the MADPL DB Server.

        Examples
        --------
        >>> output = mapdl.db._status()
        >>> print(output)
        >ENTERING THE SERVER MODE: STATUS

         DB Server is NOT currently running ..
        """
        # Need to use the health check here
        with WithinBeginLevel(self._mapdl):
            return self._mapdl.run("/DBS,SERVER,STATUS")

    def load(self, fname, progress_bar=False):
        """
        Load a MAPDL database file in memory.

        Parameters
        ----------
        fname : str
            The file name of the database to load.

        Examples
        --------
        >>> mapdl.db.load('file.db')
        """
        self._mapdl._get_file_path(fname, progress_bar=progress_bar)
        return self._mapdl.resume(fname)

    def save(self, fname, option="ALL"):
        """
        Save the MAPDL database to disk.

        Parameters
        ----------
        fname : str
            Filename to save the database to.

        option : str
            The mode for saving the database, either:

            * "ALL" - Both the model and the solution
            * "MODEL" - Just the model
            * "SOLU" - Just the solution

        Examples
        --------
        >>> mapdl.db.save('model.db', option=)
        """
        allowed = ["ALL", "MODEL", "SOLU"]
        if option.upper() not in allowed:
            raise ValueError(f"Option must be one of the following: {allowed}")

        return self._mapdl.run(f"SAVE,{fname},,,{option}")

    def clear(self, **kwargs):
        """
        Delete everything in the MAPDL database.

        Examples
        --------
        >>> mapdl.db.clear()
        """
        return self._mapdl.run("/CLEAR,ALL")

    @property
    @check_mapdl_db_is_alive
    def nodes(self):
        """
        MAPDL database nodes interface.

        Returns
        -------
        :class:`DbNodes <ansys.mapdl.core.database.nodes.DbNodes>`

        Examples
        --------
        Create a nodes instance.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> # create nodes...
        >>> nodes = mapdl.db.nodes
        >>> print(nodes)
        MAPDL Database Nodes
            Number of nodes:          270641
            Number of selected nodes: 270641
            Maximum node number:      270641

        >>> mapdl.nsel("NONE")
        >>> print(nodes)
        MAPDL Database Nodes
            Number of nodes:          270641
            Number of selected nodes: 0
            Maximum node number:      270641

        Return the selection status and the coordinates of node 22.

        >>> nodes = mapdl.db.nodes
        >>> sel, coord = nodes.coord(22)
        >>> coord
        (1.0, 0.5, 0.0, 0.0, 0.0, 0.0)

        Return all the node indices, coordinates, and angles of all the nodes.

        >>> nodes = mapdl.db.nodes
        >>> ind, coords, angles = nodes.all_asarray()
        >>> ind
        array([     1,      2,      3, ..., 270639, 270640, 270641], dtype=int32)

        >>> coords
        array([[0.    , 1.    , 0.    ],
               [0.    , 0.    , 0.    ],
               [0.    , 0.9875, 0.    ],
               ...,
               [0.9875, 0.975 , 0.925 ],
               [0.9875, 0.975 , 0.95  ],
               [0.9875, 0.975 , 0.975 ]])

        >>> angles
        array([[0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.],
               ...,
               [0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.]])

        """
        if self._nodes is None:
            from .nodes import DbNodes  # here to avoid circular import

            self._nodes = DbNodes(self)
        return self._nodes

    @property
    @check_mapdl_db_is_alive
    def elems(self):
        """
        MAPDL database element interface.

        Returns
        -------
        :class:`DbElems <ansys.mapdl.core.database.elems.DbElems>`

        Examples
        --------
        Create a MAPDL database element instance.

        >>> from ansys.mapdl.core import launch_mapdl
        >>> mapdl = launch_mapdl()
        >>> elems = mapdl.db.elems
        >>> print(elems)
        MAPDL Database Elements
            Number of elements:          64
            Number of selected elements: 64
            Maximum element number:      64

        Return the element information of element 1.

        >>> elems = mapdl.db.elems
        >>> elem_info = elems.get(1)

        Return the nodes belonging to the element.

        >>> elem_info.nodes
        [2, 27, 37, 8]

        Return the element data.

        >>> elem_info.elmdat
        [1, 1, 1, 1, 0, 0, 14, 0, 0, 0]

        """
        if self._elems is None:
            from .elems import DbElems  # here to avoid circular import

            self._elems = DbElems(self)
        return self._elems



class WorkingPlane:
    def kwpave(
        self, p1="", p2="", p3="", p4="", p5="", p6="", p7="", p8="", p9="", **kwargs
    ):
        """Moves the working plane origin to the average location of keypoints.

        APDL Command: KWPAVE

        Parameters
        ----------
        p1, p2, p3, . . . , p9
            Keypoints used in calculation of the average.  At least one must be
            defined.  If P1 = P, graphical picking is enabled and all remaining
            command fields are ignored (valid only in the GUI).

        Notes
        -----
        Moves the origin of the working plane to the average of the specified
        keypoints.  Averaging is based on the active coordinate system.

        This command is valid in any processor.
        """
        command = f"KWPAVE,{p1},{p2},{p3},{p4},{p5},{p6},{p7},{p8},{p9}"
        return self.run(command, **kwargs)

    def kwplan(self, wn="", korig="", kxax="", kplan="", **kwargs):
        """Defines the working plane using three keypoints.

        APDL Command: KWPLAN

        Parameters
        ----------
        wn
            Window number whose viewing direction will be modified to be normal
            to the working plane (defaults to 1).  If WN is a negative value,
            the viewing direction will not be modified.  If fewer than three
            points are used, the viewing direction of window WN will be used
            instead to define the normal to the working plane.

        korig
            Keypoint number defining the origin of the working plane coordinate
            system.  If KORIG = P, graphical picking is enabled and all
            remaining command fields are ignored (valid only in the GUI).

        kxax
            Keypoint number defining the x-axis orientation (defaults to the
            x-axis being parallel to the global X-axis; or if the normal to the
            working plane is parallel to the global X-axis, then defaults to
            being parallel to the global Y-axis).

        kplan
            Keypoint number defining the working plane (the normal defaults to
            the present display view [/VIEW] of window WN).

        Notes
        -----
        Defines a working plane to assist in picking operations using three
        keypoints as an alternate to the WPLANE command.  The three keypoints
        also define the working plane coordinate system.  A minimum of one
        keypoint (at the working plane origin) is required.  Immediate mode may
        also be active.  See WPSTYL command to set the style of working plane
        display.

        This command is valid in any processor.
        """
        command = f"KWPLAN,{wn},{korig},{kxax},{kplan}"
        return self.run(command, **kwargs)

    def lwplan(self, wn="", nl1="", ratio="", **kwargs):
        """Defines the working plane normal to a location on a line.

        APDL Command: LWPLAN

        Parameters
        ----------
        wn
            Window number whose viewing direction will be modified to be normal
            to the working plane (defaults to 1).  If WN is a negative value,
            the viewing direction will not be modified.

        nl1
            Number of line to be used.  If NL1 = P, graphical picking is
            enabled and all remaining command fields are ignored (valid only in
            the GUI).

        ratio
            Location on NL1, specified as a ratio of the line length.  Must be
            between 0.0 and 1.0.  If RATIO = P, use graphical picking to
            specify location on the line.

        Notes
        -----
        Defines a working plane (to assist in picking operations) normal to a
        location on a line.  See WPSTYL command to set the style of working
        plane display.

        This command is valid in any processor.
        """
        command = f"LWPLAN,{wn},{nl1},{ratio}"
        return self.run(command, **kwargs)

    def nwpave(
        self, n1="", n2="", n3="", n4="", n5="", n6="", n7="", n8="", n9="", **kwargs
    ):
        """Moves the working plane origin to the average location of nodes.

        APDL Command: NWPAVE

        Parameters
        ----------
        n1, n2, n3, . . . , n9
            Nodes used in calculation of the average.  At least one must be
            defined.  If N1 = P, graphical picking is enabled and all remaining
            command fields are ignored (valid only in the GUI).

        Notes
        -----
        Averaging is based on the active coordinate system.

        This command is valid in any processor.
        """
        command = f"NWPAVE,{n1},{n2},{n3},{n4},{n5},{n6},{n7},{n8},{n9}"
        return self.run(command, **kwargs)

    def nwplan(self, wn="", norig="", nxax="", nplan="", **kwargs):
        """Defines the working plane using three nodes.

        APDL Command: NWPLAN

        Parameters
        ----------
        wn
            Window number whose viewing direction will be modified to be normal
            to the working plane (defaults to 1).  If WN is a negative value,
            the viewing direction will not be modified.  If fewer than three
            points are used, the viewing direction of window WN will be used
            instead to define the normal to the working plane.

        norig
            Node number defining the origin of the working plane coordinate
            system.  If NORIG = P, graphical picking is enabled and all
            remaining command fields are ignored (valid only in the GUI).

        nxax
            Node number defining the x-axis orientation (defaults to the x-axis
            being parallel to the global X-axis; or if the normal to the
            working plane is parallel to the global X-axis, then defaults to
            being parallel to the global Y-axis).

        nplan
            Node number defining the working plane (the normal defaults to the
            present display view [/VIEW] of window WN).

        Notes
        -----
        Defines a working plane to assist in picking operations using three
        nodes as an alternate to the WPLANE command.  The three nodes also
        define the working plane coordinate system.  A minimum of one node (at
        the working plane origin) is required.  Immediate mode may also be
        active.  See the WPSTYL command to set the style of the working plane
        display.

        This command is valid in any processor.
        """
        command = f"NWPLAN,{wn},{norig},{nxax},{nplan}"
        return self.run(command, **kwargs)

    def wpave(
        self, x1="", y1="", z1="", x2="", y2="", z2="", x3="", y3="", z3="", **kwargs
    ):
        """Moves the working plane origin to the average of specified points.

        APDL Command: WPAVE

        Parameters
        ----------
        x1, y1, z1
            Coordinates (in the active coordinate system) of the first point.
            If X1 = P, graphical picking is enabled and all remaining command
            fields are ignored (valid only in the GUI).

        x2, y2, z2
            Coordinates (in the active coordinate system) of the second point.

        x3, y3, z3
            Coordinates (in the active coordinate system) of the third point.

        Notes
        -----
        Moves the origin of the working plane to the average of the specified
        points.  A point is considered specified only if at least one of its
        coordinates is non-blank, and at least one point (1, 2, or 3) must be
        specified.  Blank coordinates of a specified point are assumed to be
        zero.  Averaging is based on the active coordinate system.

        This command is valid in any processor.
        """
        command = f"WPAVE,{x1},{y1},{z1},{x2},{y2},{z2},{x3},{y3},{z3}"
        return self.run(command, **kwargs)

    def wpcsys(self, wn="", kcn="", **kwargs):
        """Defines the working plane location based on a coordinate system.

        APDL Command: WPCSYS

        Parameters
        ----------
        wn
            Window number whose viewing direction will be modified to be normal
            to the working plane (defaults to 1).  If WN is a negative value,
            the viewing direction will not be modified.

        kcn
            Coordinate system number.  KCN may be 0,1,2 or any previously
            defined local coordinate system number (defaults to the active
            system).

        Notes
        -----
        Defines a working plane location and orientation based on an existing
        coordinate system.  If a Cartesian system is used as the basis (KCN)
        for the working plane, the working plane will also be Cartesian, in the
        X-Y plane of the base system.  If a cylindrical, spherical, or toroidal
        base system is used, the working plane will be a polar system in the
        R-θ plane of the base system.

        If working plane tracking has been activated (CSYS,WP or CSYS,4), the
        updated active coordinate system will be of a similar type, except that
        a toroidal system will be updated to a cylindrical system.  See the
        Modeling and Meshing Guide for more information on working plane
        tracking.

        This command is valid in any processor.

        Some primitive generation commands will not honor R-theta
        transformations for non-cartesian coordinate systems. Refer to the
        primitive commands table for more information.
        """
        command = f"WPCSYS,{wn},{kcn}"
        return self.run(command, **kwargs)

    def wplane(
        self,
        wn="",
        xorig="",
        yorig="",
        zorig="",
        xxax="",
        yxax="",
        zxax="",
        xplan="",
        yplan="",
        zplan="",
        **kwargs,
    ):
        """Defines a working plane to assist in picking operations.

        APDL Command: WPLANE

        Parameters
        ----------
        wn
            Window number whose viewing direction will be modified to be normal
            to the working plane (defaults to 1).  If WN is a negative value,
            the viewing direction will not be modified.  If fewer than three
            points are used, the viewing direction of window WN will be used
            instead to define the normal to the working plane.

        xorig, yorig, zorig
            Global Cartesian coordinates of the origin of the working plane
            coordinate system.

        xxax, yxax, zxax
            Global Cartesian coordinates of a point defining the x-axis
            orientation.  The x-axis aligns with the projection of the line
            from this orientation point to the origin.

        xplan, yplan, zplan
            Global Cartesian coordinates of the third point defining the
            working plane.  This point will also define the location of the
            positive XY-sector of the working plane coordinate system.

        Notes
        -----
        Defines a working plane to assist in picking operations using the
        coordinates of three noncolinear points.  The three points also define
        the working plane coordinate system.  A minimum of one point (the
        working plane origin) is required.  Immediate mode may also be active.
        See WPSTYL command to set the style of working plane display.

        This command is valid in any processor.
        """
        command = f"WPLANE,{wn},{xorig},{yorig},{zorig},{xxax},{yxax},{zxax},{xplan},{yplan},{zplan}"
        return self.run(command, **kwargs)

    def wpoffs(self, xoff="", yoff="", zoff="", **kwargs):
        """Offsets the working plane.

        APDL Command: WPOFFS

        Parameters
        ----------
        xoff, yoff, zoff
            Offset increments defined in the working plane coordinate system.
            If only ZOFF is used, the working plane will be redefined parallel
            to the present plane and offset by ZOFF.

        Notes
        -----
        Changes the origin of the working plane by translating the working
        plane along its coordinate system axes.

        This command is valid in any processor.
        """
        command = f"WPOFFS,{xoff},{yoff},{zoff}"
        return self.run(command, **kwargs)

    def wprota(self, thxy="", thyz="", thzx="", **kwargs):
        """Rotates the working plane.

        APDL Command: WPROTA

        Parameters
        ----------
        thxy
            First rotation about the working plane Z axis (positive X toward
            Y).

        thyz
            Second rotation about working plane X axis (positive Y toward Z).

        thzx
            Third rotation about working plane Y axis (positive Z toward X).

        Notes
        -----
        The specified angles (in degrees) are relative to the orientation of
        the working plane.

        This command is valid in any processor.
        """
        command = f"WPROTA,{thxy},{thyz},{thzx}"
        return self.run(command, **kwargs)

    def wpstyl(
        self,
        snap="",
        grspac="",
        grmin="",
        grmax="",
        wptol="",
        wpctyp="",
        grtype="",
        wpvis="",
        snapang="",
        **kwargs,
    ):
        """Controls the display and style of the working plane.

        APDL Command: WPSTYL

        Parameters
        ----------
        snap
            Snap increment for a locational pick (1E-6 minimum).  If -1, turn
            off snap capability.  For example, a picked location of 1.2456 with
            a snap of 0.1 gives 1.2, with 0.01 gives 1.25, with 0.001 gives
            1.246, and with 0.025 gives 1.250 (defaults to 0.05).

        grspac
            Graphical spacing between grid points.  For graphical
            representation only and not related to snap points  (defaults to
            0.1).

        grmin, grmax
            Defines the size of a square grid (if WPCTYP = 0) to be displayed
            over a portion of the working plane.  The opposite corners of the
            grid will be located at grid points nearest the working plane
            coordinates of (GRMIN,GRMIN) and (GRMAX,GRMAX).  If a polar system
            (WPCTYP = 1), GRMAX is the outside radius of grid and GRMIN is
            ignored.  If GRMIN = GRMAX, no grid will be displayed (defaults to
            -1.0 and 1.0 for GRMIN and GRMAX respectively).

        wptol
            The tolerance that an entity's location can deviate from the
            specified working plane, while still being considered on the plane.
            Used only for locational picking of vertices for polygons and
            prisms (defaults to 0.003).

        wpctyp
            Working plane coordinate system type:

            0 - Cartesian (default).  If working plane tracking is on [CSYS,4], the updated
                active coordinate system will also be Cartesian.

            1 - Polar.  If working plane tracking is on, the updated active coordinate system
                will be cylindrical.

            2 - Polar.  If working plane tracking is on, the updated active coordinate system
                will be spherical.

        grtype
            Grid type:

            0 - Grid and WP triad.

            1 - Grid only.

            2 - WP triad only (default).

        wpvis
            Grid visibility:

            0 - Do not show GRTYPE entities (grid and/or triad) (default).

            1 - Show GRTYPE entities.  Cartesian working planes will be displayed with a
                Cartesian grid, polar with a polar grid.

        snapang
            Snap angle (0--180) in degrees.  Used only if WPCTYP = 1 or 2.
            Defaults to 5 degrees.

        Notes
        -----
        Use WPSTYL,DEFA to reset the working plane to its default location and
        style.  Use WPSTYL,STAT to list the status of the working plane.  Blank
        fields will keep present settings.

        It is possible to specify SNAP and WPTOL values that will cause
        conflicts during picking operations. Check your values carefully, and
        if problems are noted, revert to the default values.

        WPSTYL with no arguments will toggle the grid on and off.  The working
        plane can be displayed in the non-GUI interactive mode only after
        issuing a /PLOPTS,WP,1 command.  See the Modeling and Meshing Guide for
        more information on working plane tracking.  See /PLOPTS command for
        control of hidden line working plane.

        This command is valid in any processor.
        """
        command = f"WPSTYL,{snap},{grspac},{grmin},{grmax},{wptol},{wpctyp},{grtype},{wpvis},{snapang}"
        return self.run(command, **kwargs)



"""These commands may be run by PyMapdl, but are not supported from a
user-context.

"""


class _Hidden:
    def _batch(self, lab="", **kwargs):
        """APDL Command: /BATCH

        Sets the program mode to "batch."

        Parameters
        ----------
        lab
            Specifies listing mode during a batch run:

            LIST - The batch output will include a listing of the input file.

            (blank) - Suppress input data listing.

        Notes
        -----
        Sets the program mode to "batch" when included as the first line on an
        input file of ANSYS commands.  For convenience, this command is
        automatically recorded on the log file (Jobname.LOG) at the beginning
        of an interactive ANSYS session so that the log file can be reused
        later for batch input.

        Caution:: : This command  should not be entered directly in an
        interactive ANSYS session since all subsequent keyboard input is simply
        copied to a file, without further processing or prompts from the
        program (use the "system break" to exit the ANSYS program if this
        occurs).

        The recommended method for choosing batch mode, rather than using the
        /BATCH command, is to select the Batch simulation environment from the
        ANSYS Product Launcher  task in the ANSYS launcher, or the batch mode
        entry option on the ANSYS execution command when entering the program.

        This command is valid only at the Begin Level.
        """
        command = "/BATCH,%s" % (str(lab))
        return self.run(command, **kwargs)

    def _output(self, fname="", ext="", loc="", **kwargs):
        """Redirects text output to a file or to the screen.

        APDL Command: /OUTPUT

        Parameters
        ----------
        fname
            Filename and directory path (248 character maximum, including
            directory) to which text output will be redirected (defaults to
            Jobname if Ext is specified).  For interactive runs, Fname = TERM
            (or blank) redirects output to the screen.  For batch runs, Fname =
            blank (with all remaining command arguments blank) redirects output
            to the  default system output file.

        ext
            Filename extension (eight-character maximum).

        loc
            Location within a file to which output will be written:

            (blank) - Output is written starting at the top of the file (default).

            APPEND - Output is appended to the existing file.

        Notes
        -----
        Text output includes responses to every command and GUI
        function, notes, warnings, errors, and other informational
        messages.  Upon execution of /OUTPUT,Fname, Ext, ..., all
        subsequent text output is redirected to the file Fname.Ext.
        To redirect output back to the default location, issue /OUTPUT
        (no arguments).

        This command is valid in any processor.
        """
        if loc:
            return self.run(f"/OUTPUT,{fname},{ext},,{loc}", **kwargs)
        else:
            return self.run(f"/OUTPUT,{fname},{ext}", **kwargs)



"""
These DATABASE commands are generated by the GUI when picking
operations are performed.  They can also be used to add additional
parameters to a command.  For example, more points to a spline than
would be normally supported.
"""


class Picking:
    def fitem(self, nfield="", item="", itemy="", itemz="", **kwargs):
        """Identifies items chosen by a picking operation (GUI).

        APDL Command: FITEM

        Parameters
        ----------
        nfield
            Field number on the command which uses the picking data.  (Count
            the command name as a field, so that a 2 indicates the first
            command argument, 3 the second command argument, etc.)  The
            corresponding field on the command will have a P51X label.

        item
            Entity number of the entity picked.  Negative entity numbers are
            used to indicate a range of entities.  If the item picked is a
            coordinate location, then this field represents the X-coordinate.
            See also the FLST command.

        itemy, itemz
            Y and Z coordinates of a picked coordinate location.  ITEM
            represents the X coordinate.  See also the FLST  command.

        Notes
        -----
        This is a command generated by the GUI and will appear in the log file
        (Jobname.LOG) if graphical picking is used.  This command is not
        intended to be typed in directly in an ANSYS session (although it can
        be included in an input file for batch input or for use with the /INPUT
        command).

        On the log file, a set of FITEM commands is preceded by one FLST
        command which defines the picking specifications for that pick
        operation.  The data listed in the FITEM commands are used by the first
        subsequent command containing a P51X label in one of its fields.

        Caution:: : For a given entity type, a list containing an ITEM that is
        larger than the maximum defined entity, could deplete the system memory
        and produce unpredictable results.

        This command is valid in any processor.
        """
        command = "FITEM,%s,%s,%s,%s" % (str(nfield), str(item), str(itemy), str(itemz))
        return self.run(command, **kwargs)

    def flst(self, nfield="", narg="", type_="", otype="", leng="", **kwargs):
        """Specifies data required for a picking operation (GUI).

        APDL Command: FLST

        Parameters
        ----------
        nfield
            Field number on the command which uses the picking data.  (Count
            the command name as a field, so that a 2 indicates the first
            command argument, 3 for the second command argument, etc.)  The
            corresponding field on the command will have a P51X label.

        narg
            Number of items in the picked list.

        type\_
            Type of items picked:

            1 - Node numbers

            2 - Element numbers

            3 - Keypoint numbers

            4 - Line numbers

            5 - Area numbers

            6 - Volume numbers

            7 - Trace points

            8 - Coordinate locations (in Global Cartesian coordinates)

            9 - Screen picks (in X, Y screen coordinates (-1 to 1))

        otype
            Data order:

            NOOR - Data is not ordered (default).

            ORDER - Data is in an ordered list (such as for the E,P51X and A,P51X commands, in
                    which the order of the data items is significant for the
                    picking operation).

        leng
            Length of number of items describing the list (should equal NARG if
            Otype = NOOR; default).

        Notes
        -----
        Specifies data required for the FITEM command during a picking
        operation.  This is a command generated by the GUI and will appear in
        the log file (Jobname.LOG) if graphical picking is used.  This command
        is not intended to be typed in directly in an ANSYS session (although
        it can be included in an input file for batch input or for use with the
        /INPUT command).

        On the log file, FLST will always be followed by one or more FITEM
        commands which in turn are followed by the ANSYS command that contains
        a P51X label in one of its fields. This set of commands should not be
        edited.

        This command is valid in any processor.
        """
        command = "FLST,%s,%s,%s,%s,%s" % (
            str(nfield),
            str(narg),
            str(type_),
            str(otype),
            str(leng),
        )
        return self.run(command, **kwargs)



class MultiFieldSolverGlobalControls:
    def mfanalysis(self, key="", **kwargs):
        """Activates or deactivates an ANSYS Multi-field solver analysis.

        APDL Command: MFANALYSIS

        Parameters
        ----------
        key
            Multifield analysis key:

            ON - Activates an ANSYS Multi-field solver analysis.

            OFF - Deactivates an ANSYS Multi-field solver analysis (default).

        Notes
        -----
        This command is also valid in PREP7.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFANALYSIS,{key}"
        return self.run(command, **kwargs)

    def mfclear(self, option="", value="", **kwargs):
        """Deletes ANSYS Multi-field solver analysis settings.

        APDL Command: MFCLEAR

        Parameters
        ----------
        option
            SOLU

            SOLU - Resets all ANSYS solution commands except KBC to their default states. This
                   option clears analysis options when setting up different
                   fields for an ANSYS Multi-field solver analysis.

            FIELD - Deletes all ANSYS Multi-field solver specifications for the specified field
                    number.

            SINT - Deletes all ANSYS Multi-field solver specifications for the specified surface
                   interface number.

            VINT - Deletes all ANSYS Multi-field solver specifications for the volumetric
                   interface number.

            ORD - Deletes the analysis order specified by the MFORDER command.

            EXT - Deletes external fields specified by the MFEXTER command

            MFLC - Deletes load transfers specified by the MFLCOMM command

        value
            Use only for Option = FIELD, SINT, or VINT.

        Notes
        -----
        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFCLEAR,{option},{value}"
        return self.run(command, **kwargs)

    def mffr(self, fname="", lab="", rfini="", rfmin="", rfmax="", **kwargs):
        """Setup Multi-Field relaxation factors for field solutions.

        APDL Command: MFFR

        Parameters
        ----------
        fname
            Field name (MFX) or number (MFS). Must be the ANSYS field (cannot
            be a CFX field).

        lab
            Label name. Valid values are DISP and TEMP.

        rfini
            Initial relaxation factor. Defaults to 0.75.

        rfmin
            Minimum relaxation factor. Defaults to RFINI.

        rfmax
            Maximum relaxation factor. Defaults to RFINI.

        Notes
        -----
        Use this command to relax the field solutions in fluid-solid
        interaction analyses and thermal-thermal analyses for a better
        convergence rate in coupled problems, especially cases that need
        dynamic relaxation. The ANSYS field that has the MFFR command applied
        will do only one nonlinear stagger iteration within each multi-field
        stagger; the convergence of the ANSYS field solver will be satisfied
        through multiple multi-field staggers. Note that the CFX field solver
        can have multiple iterations within the field solver; see the CFX
        documentation for more details. ANSYS will not terminate the nonlinear
        field solution until the ANSYS field solver converges or reaches the
        maximum number of multi-field staggers as specified on MFITER.

        The interface load relaxation (MFRELAX) will be automatically turned
        off for the corresponding surface loads that have MFFR applied. The
        automatic change of the relaxation factor for accelerating the
        nonlinear convergence of the coupled field solution is based on
        Aitken's acceleration scheme.

        This command is valid only with coupled problems involving surface load
        transfer only. No subcycling is allowed for the field solver if using
        this command.
        """
        command = f"MFFR,{fname},{lab},{rfini},{rfmin},{rfmax}"
        return self.run(command, **kwargs)

    def mfinter(self, option="", **kwargs):
        """Specifies the interface load transfer interpolation option for an ANSYS

        APDL Command: MFINTER
        Multi-field solver analysis.

        Parameters
        ----------
        option
            Interface load transfer option:

            CONS - Conservative formulation for load transfer.

            NONC - Nonconservative formulation for load transfer (default).

        Notes
        -----
        This command only applies to the interpolation method for forces, heat
        flux, and heat generation. Displacement and temperature transfers are
        always nonconservative.

        For more information on conservative and nonconservative load transfer,
        see Load Transfer in the Coupled-Field Analysis Guide.

        This command is also valid in PREP7.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFINTER,{option}"
        return self.run(command, **kwargs)

    def mflist(self, option="", value="", **kwargs):
        """Lists the settings for an ANSYS Multi-field solver analysis.

        APDL Command: MFLIST

        Parameters
        ----------
        option
            ALL

            ALL - Lists all ANSYS Multi-field solver analysis options.

            SOLU - Lists all solution-related ANSYS Multi-field solver options.

            FIELD - Lists all ANSYS Multi-field solver options related to the specified field
                    number.

            SINT - Lists all surface interface information for the specified surface interface
                   number.

            VINT - Lists all volumetric interface information for the specified volumetric
                   interface number.

        value
            Use only for Option = FIELD, SINT, or VINT.

        Notes
        -----
        This command is also valid in PREP7.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFLIST,{option},{value}"
        return self.run(command, **kwargs)

    def mforder(
        self,
        fnumb1="",
        fnumb2="",
        fnumb3="",
        fnumb4="",
        fnumb5="",
        fnumb6="",
        fnumb7="",
        fnumb8="",
        fnumb9="",
        fnumb10="",
        fnumb11="",
        fnumb12="",
        fnumb13="",
        fnumb14="",
        fnumb15="",
        fnumb16="",
        fnumb17="",
        fnumb18="",
        fnumb19="",
        fnumb20="",
        **kwargs,
    ):
        """Specifies field solution order for an ANSYS Multi-field solver

        APDL Command: MFORDER
        analysis.

        Parameters
        ----------
        fnumb1, fnumb2, fnumb3, . . . , fnumb20
            Field numbers defined by the MFELEM command.

        Notes
        -----
        You can define up to twenty fields in an ANSYS Multi-field solver
        analysis.

        This command is also valid in PREP7.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFORDER,{fnumb1},{fnumb2},{fnumb3},{fnumb4},{fnumb5},{fnumb6},{fnumb7},{fnumb8},{fnumb9},{fnumb10},{fnumb11},{fnumb12},{fnumb13},{fnumb14},{fnumb15},{fnumb16},{fnumb17},{fnumb18},{fnumb19},{fnumb20}"
        return self.run(command, **kwargs)

    def mfpsimul(self, gname="", fname1="", fname2="", **kwargs):
        """Sets up a field solver group to simultaneously process with code

        APDL Command: MFPSIMUL
        coupling analyses.

        Parameters
        ----------
        gname
            Sets the group name with a character string of up to 80 characters.

        fname1, fname2
            Sets the field solver 1 and field solver 2 names, which are
            processed simultaneously, with a character string of up to 80
            characters.

        Notes
        -----
        This command is used to define a group of simultaneously-processed
        field solvers in an MFX analysis. For example, to define group g1 with
        field solvers ansys-code and cfx-code, enter MFPS,g1,ansys-code,cfx-
        code.

        To indicate groups of sequentially-processed field solvers for your MFX
        analysis, create two groups (g1 and g2).

        A field solver refers to a specific instance of an ANSYS or CFX solver
        execution that is defined by the respective input file(s) referenced
        when starting the solver (through the launcher or from the command
        line). The field solver names that are referenced in several MFX
        commands must be consistent with the names that will be used when
        starting the coupled simulation.

        Note:: : When running MFX from the launcher, you must use ANSYS and CFX
        (uppercase) as the field solver names (MFPSIMUL) in your input file.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFPSIMUL,{gname},{fname1},{fname2}"
        return self.run(command, **kwargs)

    def mfsorder(self, gname1="", gname2="", **kwargs):
        """Sets up the solution sequence of simultaneous field solver groups for

        APDL Command: MFSORDER
        code coupling analyses.

        Parameters
        ----------
        gname1, gname2
            Specifies the group name for groups defined by the MFPSIMUL command
            with a character string of up to 80 characters.

        Notes
        -----
        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MFSORDER,{gname1},{gname2}"
        return self.run(command, **kwargs)

    def mfwrite(self, fname="", ext="", **kwargs):
        """Writes an ANSYS master input file for MFX multiple code coupling.

        APDL Command: MFWRITE

        Parameters
        ----------
        fname
            File name and directory path (248 characters maximum, including the
            characters needed for the directory path).  An unspecified
            directory path defaults to the working directory; in this case, you
            can use all 248 characters for the file name.

        ext
            Filename extension (eight-character maximum).

        Notes
        -----
        When working interactively, you need to issue this command as the last
        step in your setup process. This command will write out the input file
        that you will then use to submit the MFX analysis. This file will
        include the /SOLU, SOLVE, and FINISH commands.

        See Multi-field Commands in the Coupled-Field Analysis Guide for a list
        of all ANSYS Multi-field solver commands and their availability for MFS
        and MFX analyses.
        """
        command = f"MFWRITE,{fname},{ext}"
        return self.run(command, **kwargs)



class Rezoning:
    def rezone(self, option="", ldstep="", sbstep="", **kwargs):
        """Initiates the rezoning process, sets rezoning options, and rebuilds the

        APDL Command: REZONE
        database.

        Parameters
        ----------
        option
            The rezoning method to employ:

            MANUAL - Manual rezoning. You decide when to use rezoning, what region(s) to rezone, and
                     what remeshing method to use on the selected region(s).
                     This method is currently the default and only option.

        ldstep
            The load step number at which rezoning should occur. The default
            value is the highest load step number found in the Jobname.Rnnn
            files (for the current jobname and in the current directory).

        sbstep
            The substep number of the specified load step (LDSTEP) at which
            rezoning should occur. The default value is the highest substep
            number found in the specified load step in the Jobname.Rnnn files
            (for the current jobname and in the current directory).

        Notes
        -----
        The REZONE command rebuilds the database (.db file) based on the
        specified load step and substep information, and updates nodes to their
        deformed position for remeshing.

        Before issuing this command, clear the database via the /CLEAR command.

        For more information, see Rezoning in the Advanced Analysis Guide.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"REZONE,{option},{ldstep},{sbstep}"
        return self.run(command, **kwargs)

    def mapsolve(self, maxsbstep="", **kwargs):
        """Maps solved node and element solutions from an original mesh to a new

        APDL Command: MAPSOLVE
        mesh.

        Parameters
        ----------
        maxsbstep
            The maximum number of substeps for rebalancing the residuals. The
            default value is 5.

        Notes
        -----
        Used during the rezoning process, the MAPSOLVE command maps solved node
        and element solutions from the original mesh to the new mesh and
        achieves equilibrium based on the new mesh.

        Additional substeps are necessary to reduce the residuals to zero.

        During the rebalancing stage, the external loads and time remain
        unchanged.

        The MAPSOLVE command is valid only for rezoning (REZONE).

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        command = f"MAPSOLVE,{maxsbstep}"
        return self.run(command, **kwargs)

    def mapvar(
        self,
        option="",
        matid="",
        istrtstress="",
        ntenstress="",
        istrtstrain="",
        ntenstrain="",
        istrtvect="",
        nvect="",
        **kwargs,
    ):
        """Defines tensors and vectors in user-defined state variables for

        APDL Command: MAPVAR
        rezoning and in 2-D to 3-D analyses.

        Parameters
        ----------
        option
            DEFINE

            DEFINE - Define variables for the specified MatId material ID (default).

            LIST - List the defined variables for the specified MatId material ID.

        matid
            The material ID for the state variables which you are defining
            (Option = DEFINE) or listing (Option = LIST).

        istrtstress
            The start position of stress-like tensors in the state variables.
            This value must be either a positive integer or 0 (meaning no
            stress-like tensors).

        ntenstress
            The number of stress-like tensors in the state variables. This
            value must be either a positive integer (or 0), and all stress-like
            tensors must be contiguous.

        istrtstrain
            The start position of strain-like tensors in the state variables.
            This value must be either a positive integer or 0 (meaning no
            strain-like tensors).

        ntenstrain
            The number of strain-like tensors in the state variables. This
            value must be either a positive integer (or 0), and all strain-like
            tensors must be contiguous.

        istrtvect
            The start position of vectors in the state variables. This value
            must be either a positive integer or 0 (meaning no vectors).

        nvect
            The number of vectors in the state variables. This value must be
            either a positive integer (or 0), and all vectors must be
            contiguous.

        Notes
        -----
        The MAPVAR command identifies the tensors and vectors in user-defined
        state variables (TB,STATE) for user-defined materials (TB,USER and
        UserMat or UserMatTh) or user-defined creep laws (TB,CREEP,,,,100 and
        UserCreep).

        The command ensures that user-defined state variables are mapped
        correctly during rezoning and in 2-D to 3-D analyses.

        In a rezoning operation, MAPVAR must be issued after remeshing
        (REMESH,FINISH) but before mapping (MAPSOLVE).
        """
        command = f"MAPVAR,{option},{matid},{istrtstress},{ntenstress},{istrtstrain},{ntenstrain},{istrtvect},{nvect}"
        return self.run(command, **kwargs)

    def remesh(self, action="", filename="", ext="", opt1="", opt2="", **kwargs):
        """Specifies the starting and ending remeshing points, and other options,

        APDL Command: REMESH
        for rezoning.

        Parameters
        ----------
        action
            START

            START - Starts the remeshing operation.

            FINISH - Ends the remeshing operation.

            READ - Reads in a generic (.cdb format) new mesh file generated by a third-party
                   application. This remeshing option applies to both 2-D and
                   3-D rezoning.

            SPLIT - Splits selected elements of an existing 2-D or 3-D mesh such that a
                    quadrilateral element is split into four quadrilaterals, a
                    degenerate quadrilateral is split into three
                    quadrilaterals, and a quadratic triangular element is split
                    into four quadratic triangles.  A tetrahedral element is
                    split into eight tetrahedra.

        filename
            Name of a .cdb generic mesh file. The default value is jobname.
            Valid only when Action = READ.

        ext
            File name extension. The only valid (and the default) extension is
            CDB. Valid only when Action = READ.

        opt1
            Specifies options for the new mesh when using a generic imported
            mesh file or the mesh-splitting remeshing method. Valid only when
            Action = READ or Action = SPLIT.

            REGE - Regenerates all node and element numbers on the new
                    mesh using an offset of the highest existing node
                    and element numbers. This is the default behavior
                    when Action = READ; otherwise, this value is
                    ignored.

            KEEP - Keeps the similarly numbered nodes and elements in
                    the new and the old meshes unchanged. Valid only
                    when Action = READ.

            TRAN - Generates transition elements to ensure nodal
                    compatibility between split and unsplit parts of
                    the mesh. Valid only when Action = SPLIT for 2-D
                    analyses.

        opt2
            Specifies transition options for the mesh when elements are split.
            These options are valid only when Action = SPLIT for 2-D analyses.

            QUAD  - Minimizes the number of degenerate elements in the transition mesh and tries to
                    maximize the number of quadrilateral transition elements
                    across several layers of elements from the split regions.
                    This is the default behavior.

            DEGE  - Creates transition zones between the split and unsplit parts of the mesh using
                    mostly degenerate elements with a single element layer.

        Notes
        -----
        The REMESH command is valid only during the rezoning (REZONE) process.

        In rezoning, a REMESH,START command temporarily exits the /SOLU
        solution processor and enters a special mode of the /PREP7
        preprocessor, after which a limited number of preprocessing commands
        are available for mesh control, but no solution commands are valid.

        A REMESH,FINISH command exits the remeshing process and reenters the
        solution processor, at which point no preprocessing commands are
        available. If the new mesh exists, the command creates contact elements
        if needed, and transfers all boundary conditions (BCs) and loads from
        the original mesh to the new mesh. You can issue any list or plot
        command to verify the created contact elements, transferred BCs, and
        loads. A REMESH,FINISH command is valid only after a previously issued
        REMESH,START command, and is the only way to safely end the remeshing
        operation (and exit the special mode of the /PREP7 preprocessor).

        A REMESH,READ command is valid only when you want to perform a rezoning
        operation using a generic new mesh generated by a third-party
        application (rather than a new mesh generated internally by the ANSYS
        program). The command is valid between REMESH,START and REMESH,FINISH
        commands. In this case, the only valid file extension is .cdb (Ext =
        CDB). When Option = KEEP, ANSYS assumes that the common node and
        element numbers between the old and the new mesh are topologically
        similar (that is, these commonly numbered areas have the same element
        connectivity and nodal coordinates).

        A REMESH,SPLIT command is valid only when you wish to perform a
        rezoning operation by splitting the existing mesh. The command is valid
        between REMESH,START and REMESH,FINISH commands.

        You can use REMESH,READ and REMESH,SPLIT commands for horizontal
        multiple rezoning provided that the meshes used in REMESH,READ do not
        intersect. (ANSYS recommends against issuing an AREMESH command after
        issuing either of these commands.)

        For more detailed about the remeshing options available to you during a
        rezoning operation, see Rezoning in the Advanced Analysis Guide.

        Distributed ANSYS Restriction: This command is not supported in
        Distributed ANSYS.
        """
        return self.run(f"REMESH,{action},{filename},{ext},,{opt1},{opt2}", **kwargs)

    def aremesh(self, lcomb="", angle="", **kwargs):
        """Generates an area in which to create a new mesh for rezoning.

        APDL Command: AREMESH

        Parameters
        ----------
        lcomb
            Specifies how to combine adjacent line segments:

             0 - Line segments combined by connecting ends to ends. This value is the default.

            -1 - No line segments combined.

        angle
            The maximum angle (in degrees) allowed for connecting two line
            segments together. The default value is 30. This value is valid
            only when LCOMB = 0.

        Notes
        -----
        Issue the AREMESH command after issuing a REMESH,START command and
        before issuing a REMESH,FINISH command.

        The AREMESH command cannot account for an open area (or "hole") inside
        a completely enclosed region. Instead, try meshing around an open area
        by selecting two adjoining regions; for more information, see Hints for
        Remeshing Multiple Regions .
        """
        command = f"AREMESH,{lcomb},{angle}"
        return self.run(command, **kwargs)



"""pymapdl examples"""
import os

from matplotlib.colors import ListedColormap
import numpy as np

# get location of this folder and the example files
dir_path = os.path.dirname(os.path.realpath(__file__))

# add any files you'd like to import here.  For example:
wing_model = os.path.join(dir_path, "wing.dat")

# be sure to add the input file directly in this directory
# This way, files can be loaded with:
# from ansys.mapdl.core import examples
# examples.wing_model


def ansys_colormap():
    """Return the default ansys color map made of 9 colours (blue-green-red)."""
    colors = (
        np.array(
            [
                [0, 0, 255],
                [0, 178, 255],
                [0, 255, 255],
                [0, 255, 178],
                [0, 255, 0],
                [178, 255, 0],
                [255, 255, 0],
                [255, 178, 0],
                [255, 0, 0],
            ],
            dtype=float,
        )
        / 255
    )
    return ListedColormap(colors)
